/***
*
* BNBT Beta 8.0 - A C++ BitTorrent Tracker
* Copyright (C) 2003-2004 Trevor Hogan
*
* This library is free software; you can redistribute it and/or
* modify it under the terms of the GNU Lesser General Public
* License as published by the Free Software Foundation; either
* version 2.1 of the License, or (at your option) any later version.
*
* This library is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
* Lesser General Public License for more details.
*
* You should have received a copy of the GNU Lesser General Public
* License along with this library; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*
***/

#include "bnbt.h"
#include "atom.h"
#include "bencode.h"
#include "config.h"
#include "html.h"
#include "link.h"
#include "server.h"
#include "tracker.h"
#include "util.h"

void CTracker :: serverResponseAdmin( struct request_t *pRequest, struct response_t *pResponse )
{
	// Set the start time
	const struct bnbttv btv( UTIL_CurrentTime( ) );

	// Verify that the IP is permitted to access the tracker
	if( m_ucIPBanMode != 0)
		if( IsIPBanned( pRequest, pResponse, btv, gmapLANG_CFG["admin_page"], string( CSS_ADMIN ), NOT_INDEX ) )
			return;

	// User must have administration rights
	if( pRequest->user.ucAccess & ACCESS_ADMIN )
	{
		//
		// kill tracker
		//

		if( pRequest->mapParams["ap_kill"] == "1" )
		{
			UTIL_LogPrint( "Admin: Kill Tracker\n" );

			gpServer->Kill( );

			return;
		}

		//
		// count unique peers
		//

		if( m_bCountUniquePeers && pRequest->mapParams["ap_recount"] == "1" )
		{
			UTIL_LogPrint( "Admin: Count Unique Peers\n" );

			gpServer->getTracker( )->CountUniquePeers( );

			UTIL_LogPrint( "Admin: %i peers\n", m_pIPs->getValuePtr( )->size( ) );

			return JS_ReturnToPage( pRequest, pResponse, ADMIN_HTML );
		}

		//
		// reset tracker link
		//

		if( pRequest->mapParams["ap_relink"] == "1" )
		{
			UTIL_LogPrint( "Admin: Reset Tracker Link\n" );

			if( gpLink )
			{
				// Resetting tracker link
				gpLink->Kill( );
	
				return JS_ReturnToPage( pRequest, pResponse, ADMIN_HTML );
			}
			else
			{
				// This tracker does not own a tracker link (it might be a tracker hub)
				m_bFlagNotOwnLinkAlert = true;

				UTIL_LogPrint( "Admin: This tracker does not own a tracker link\n" );

				return JS_ReturnToPage( pRequest, pResponse, ADMIN_HTML );
			}
		}  

		//
		// reset tracker hub link
		//

		if( pRequest->mapParams["ap_relinkhub"] == "1" )
		{
			UTIL_LogPrint( "Admin: Reset Tracker Hub Link\n" );

			if( gpHUBLink )
			{
				// Resetting tracker link
				gpHUBLink->Kill( );
	
				return JS_ReturnToPage( pRequest, pResponse, ADMIN_HTML );
			}
			else
			{
				// This tracker does not own a tracker hub link (it might be a tracker hub)
				m_bFlagNotOwnHUBLinkAlert = true;

				UTIL_LogPrint( "Admin: This tracker does not own a tracker hub link\n" );

				return JS_ReturnToPage( pRequest, pResponse, ADMIN_HTML );
			}
		}  
			
		//
		// Dump XML
		//

		if( !xmldump.strName.empty( ) && pRequest->mapParams["ap_xml"] == "1" )
		{
			UTIL_LogPrint( "Admin: Update XML\n" );

			saveXML( );

			m_ulDumpXMLNext = GetTime( ) + m_uiDumpXMLInterval;

			m_bFlagXMLAlert = true;

			return JS_ReturnToPage( pRequest, pResponse, ADMIN_HTML );
		}

		if( !xmldump.strName.empty( ) )
		{
			if( !xmldump.strDir.empty( ) )
				xmldump.strFile = UTIL_ReadFile( ( xmldump.strDir + RESPONSE_STR_SEPERATOR + xmldump.strName ).c_str( ) );
			else
				xmldump.strFile = UTIL_ReadFile( xmldump.strName.c_str( ) );
		}
		else
			UTIL_LogPrint( "Admin: Update XML - file name not set\n" );

		//
		// Dump Dynstat
		//

#if defined ( XBNBT_GD )
		if( m_bDynstatGenerate && !m_strDynstatFontFile.empty( ) && !m_strDynstatLinkURL.empty( ) && pRequest->mapParams["ap_dynstat"] == "1" )
		{
			UTIL_LogPrint( "Admin: Update Dynstat Images\n" );
			
			runGenerateDynstat( );

			return JS_ReturnToPage( pRequest, pResponse, ADMIN_HTML );
		}
#endif

		//
		// Dump RSS (thanks labarks)
		//

		if( !rssdump.strName.empty( ) && pRequest->mapParams["ap_rss"] == "1" )
		{
			UTIL_LogPrint( "Admin: Update RSS\n" );
			
			if( m_ucDumpRSSFileMode == 0 || m_ucDumpRSSFileMode == 2 )
			{
				if( gbDebug )
					if( gucDebugLevel & DEBUG_TRACKER )
						UTIL_LogPrint( ( gmapLANG_CFG["dumping_rss"] + "\n" ).c_str( ) );

				saveRSS( );
			}

			if( m_ucDumpRSSFileMode == 1 || m_ucDumpRSSFileMode == 2 )
			{
				string strTag = string( );

				for( vector< pair<string, string> > :: iterator it = m_vecTags.begin( ); it != m_vecTags.end( ); it++ )
				{
					strTag = (string)(*it).first;

					if( gbDebug )
						if( gucDebugLevel & DEBUG_TRACKER )
							UTIL_LogPrint( ( gmapLANG_CFG["dumping_rss_for"] + "\n" ).c_str( ), strTag.c_str( ) );

					saveRSS( strTag );
				}

				if( !m_vecTags.size( ) && m_ucDumpRSSFileMode == 1 )
					UTIL_LogPrint( ( gmapLANG_CFG["dumping_rss_tags_warning"] + "\n" ).c_str( ) );
			}

			m_ulDumpRSSNext = GetTime( ) + m_uiDumpRSSInterval * 60;

			if( !rssdump.strName.empty( ) )
			{
				if( !rssdump.strDir.empty( ) )
					rssdump.strFile = UTIL_ReadFile( ( rssdump.strDir + RESPONSE_STR_SEPERATOR + rssdump.strName ).c_str( ) );
				else
					rssdump.strFile = UTIL_ReadFile( rssdump.strName.c_str( ) );
			}
			else
				UTIL_LogPrint( "Admin: Update RSS - file name not set\n" );

			m_bFlagRSSAlert = true;

			return JS_ReturnToPage( pRequest, pResponse, ADMIN_HTML );
		}

		//
		// display the admin page
		//

		// Output common HTML head
		HTML_Common_Begin( pRequest, pResponse, gmapLANG_CFG["admin_page"], string( CSS_ADMIN ), string( ), NOT_INDEX, CODE_200 );

		// JavaScript
		pResponse->strContent += "<script type=\"text/javascript\">\n";
		pResponse->strContent += "<!--\n";

		if( m_bCountUniquePeers )
		{
			pResponse->strContent += "function admin_count() {\n";
			pResponse->strContent += "window.location=\"" + RESPONSE_STR_ADMIN_HTML + "?ap_recount=1\"\n";
			pResponse->strContent += "}\n\n";
		}

		if( !rssdump.strName.empty( ) )
		{
			pResponse->strContent += "function admin_rss() {\n";
			pResponse->strContent += "window.location=\"" + RESPONSE_STR_ADMIN_HTML + "?ap_rss=1\"\n";
			pResponse->strContent += "}\n\n";
		}

		if( !xmldump.strName.empty( ) )
		{
			pResponse->strContent += "function admin_xml() {\n";
			pResponse->strContent += "window.location=\"" + RESPONSE_STR_ADMIN_HTML + "?ap_xml=1\"\n";
			pResponse->strContent += "}\n\n";
		}

#if defined ( XBNBT_GD )
		if( m_bDynstatGenerate && !m_strDynstatFontFile.empty( ) && !m_strDynstatLinkURL.empty( ) )
		{
			pResponse->strContent += "function admin_dynstat() {\n";
			pResponse->strContent += "window.location=\"" + RESPONSE_STR_ADMIN_HTML + "?ap_dynstat=1\"\n";
			pResponse->strContent += "}\n\n";
		}
#endif

		pResponse->strContent += "function kill_confirm()\n";
		pResponse->strContent += "{\n";
		pResponse->strContent += "var name=confirm(\"" + gmapLANG_CFG["js_kill_the_tracker_q"] + "\")\n";
		pResponse->strContent += "if (name==true)\n";
		pResponse->strContent += "{\n";
		pResponse->strContent += "window.location=\"" + RESPONSE_STR_ADMIN_HTML + "?ap_kill=1\"\n";
		pResponse->strContent += "}\n";
		pResponse->strContent += "}\n\n";

		pResponse->strContent += "function relink_confirm()\n";
		pResponse->strContent += "{\n";
		pResponse->strContent += "var name=confirm(\"" + gmapLANG_CFG["js_reset_tracker_link_q"] + "\")\n";
		pResponse->strContent += "if (name==true)\n";
		pResponse->strContent += "{\n";
		pResponse->strContent += "window.location=\"" + RESPONSE_STR_ADMIN_HTML + "?ap_relink=1\"\n";
		pResponse->strContent += "}\n";
		pResponse->strContent += "}\n\n";

		pResponse->strContent += "function relinkhub_confirm()\n";
		pResponse->strContent += "{\n";
		pResponse->strContent += "var name=confirm(\"" + gmapLANG_CFG["js_reset_tracker_hub_link_q"] + "\")\n";
		pResponse->strContent += "if (name==true)\n";
		pResponse->strContent += "{\n";
		pResponse->strContent += "window.location=\"" + RESPONSE_STR_ADMIN_HTML + "?ap_relinkhub=1\"\n";
		pResponse->strContent += "}\n";
		pResponse->strContent += "}\n\n";

		pResponse->strContent += "//-->\n";
		pResponse->strContent += "</script>\n\n";

		//
		// tracker links
		//

		pResponse->strContent += "<div class=\"admin_table\">\n";
		pResponse->strContent += "<table summary=\"tlink\">\n";
		
		if( gpLink )
		{
			pResponse->strContent += "<tr><th colspan=1>" + gmapLANG_CFG["admin_tracker_links"] + "</th>\n";
			pResponse->strContent += "<th colspan=1>";
			pResponse->strContent += Button_JS_Link( "confirm_relink", gmapLANG_CFG["reset"], "relink_confirm()" );
			pResponse->strContent += "</th></tr>\n";
			pResponse->strContent += "<tr><th>" + gmapLANG_CFG["established"] + "</th>";
			pResponse->strContent += "<td class=\"tracker_connection\">" + gtXStats.date.sLinkEstablished + "</td></tr>\n";
		}
		else
		{
			pResponse->strContent += "<tr><th colspan=2>" + gmapLANG_CFG["admin_tracker_links"] + "</th></tr>\n";
		}

		pResponse->strContent += "<tr><th>" + gmapLANG_CFG["type"] + "</th>";

		if( gpLinkServer )
		{
			pResponse->strContent += "<td class=\"tracker_link_type\">" + gmapLANG_CFG["admin_primary_tracker"] + "</td></tr>\n";
			pResponse->strContent += "<tr><th>" + gmapLANG_CFG["admin_connections"] + "</th><td class=\"tracker_connection\">";

			gpLinkServer->m_mtxLinks.Claim( );

			for( vector<CLinkClient *> :: iterator it = gpLinkServer->m_vecLinks.begin( ); it != gpLinkServer->m_vecLinks.end( ); it++ )
			{
				pResponse->strContent += (*it)->getName( );

				if( (*it)->m_bActive )
					pResponse->strContent += " " + gmapLANG_CFG["admin_active"];
				else
					pResponse->strContent += " " + gmapLANG_CFG["admin_not_active"];

				if( it + 1 != gpLinkServer->m_vecLinks.end( ) )
					pResponse->strContent += "<br>";
			}

			gpLinkServer->m_mtxLinks.Release( );

			pResponse->strContent += "</td></tr>\n";
		}
		else if( gpLink )
		{
			pResponse->strContent += "<td class=\"tracker_link_type\">" + gmapLANG_CFG["admin_secondary_tracker"] + "</td></tr>\n";
			pResponse->strContent += "<tr><th>" + gmapLANG_CFG["admin_connections"] + "</th><td class=\"tracker_connection\">" + gpLink->getName( ) + "</td></tr>\n";
		}
		else
			pResponse->strContent += "<td class=\"tracker_connection\">" + gmapLANG_CFG["admin_no_link"] + "</td></tr>\n";

		pResponse->strContent += "</table>\n</div>\n";

		//
		// tracker hub links
		//

		pResponse->strContent += "<div class=\"admin_table\">\n";
		pResponse->strContent += "<table summary=\"tlink\">\n";
		
		if( gpHUBLink )
		{
			pResponse->strContent += "<tr><th colspan=1>" + gmapLANG_CFG["admin_tracker_hub_links"] + "</th>\n";
			pResponse->strContent += "<th colspan=1>";
			pResponse->strContent += Button_JS_Link( "confirm_relink", gmapLANG_CFG["reset"], "relinkhub_confirm()" );
			pResponse->strContent += "</th></tr>\n";

			pResponse->strContent += "<tr><th>" + gmapLANG_CFG["established"] + "</th>";
			pResponse->strContent += "<td class=\"tracker_connection\">" + gtXStats.date.sHubLinkEstablished + "</td></tr>\n";
		}
		else
		{
			pResponse->strContent += "<tr><th colspan=2>" + gmapLANG_CFG["admin_tracker_hub_links"] + "</th></tr>\n";
		}

		pResponse->strContent += "<tr><th>" + gmapLANG_CFG["type"] + "</th>";

		if( gpHUBLinkServer )
		{
			pResponse->strContent += "<td class=\"tracker_link_type\">" + gmapLANG_CFG["admin_primary_tracker"] + "</td></tr>\n";
			pResponse->strContent += "<tr><th>" + gmapLANG_CFG["admin_connections"] + "</th><td class=\"tracker_connection\">";

			gpHUBLinkServer->m_mtxLinks.Claim( );

			for( vector<CHUBLinkClient *> :: iterator it = gpHUBLinkServer->m_vecLinks.begin( ); it != gpHUBLinkServer->m_vecLinks.end( ); it++ )
			{
				pResponse->strContent += (*it)->getName( );

				if( (*it)->m_bActive )
					pResponse->strContent += " " + gmapLANG_CFG["admin_active"];
				else
					pResponse->strContent += " " + gmapLANG_CFG["admin_not_active"];

				if( it + 1 != gpHUBLinkServer->m_vecLinks.end( ) )
					pResponse->strContent += "<br>";
			}

			gpHUBLinkServer->m_mtxLinks.Release( );

			pResponse->strContent += "</td></tr>\n";
		}
		else if( gpHUBLink )
		{
			pResponse->strContent += "<td class=\"tracker_link_type\">" + gmapLANG_CFG["admin_secondary_tracker"] + "</td></tr>\n";
			pResponse->strContent += "<tr><th>" + gmapLANG_CFG["admin_connections"] + "</th><td class=\"tracker_connection\">" + gpHUBLink->getName( ) + "</td></tr>\n";
		}
		else
			pResponse->strContent += "<td class=\"tracker_connection\">" + gmapLANG_CFG["admin_no_link"] + "</td></tr>\n";

		pResponse->strContent += "</table>\n</div>\n";

		//
		// clients
		//

		const unsigned int cuiOnline( ( unsigned int )gpServer->m_vecClients.size( ) );

		pResponse->strContent += "<p class=\"users_online\">" + gmapLANG_CFG["admin_serving"] + " " + CAtomInt( cuiOnline ).toString( ) + " ";

		if ( cuiOnline > 1 )
			pResponse->strContent += gmapLANG_CFG["admin_serving_clients"];
		else
			pResponse->strContent += gmapLANG_CFG["admin_serving_client"];

		pResponse->strContent += "</p>\n";

		// Button List
		pResponse->strContent += "<p class=\"button_list\">";

		// count unique peers
		if( m_bCountUniquePeers )
			pResponse->strContent += Button_JS_Link( "link_count", gmapLANG_CFG["admin_count_peers"] + " (" + CAtomLong( m_pIPs->getValuePtr( )->size( ) ).toString( ) + ")", "admin_count()" );

		// RSS
		// addition by labarks
		if( !rssdump.strName.empty( ) )
			pResponse->strContent += Button_JS_Link( "link_rss", gmapLANG_CFG["admin_update_rss"], "admin_rss()" );

		// XML
		if( !rssdump.strName.empty( ) )
			pResponse->strContent += Button_JS_Link( "link_xml", gmapLANG_CFG["admin_update_xml"], "admin_xml()" );

		// Dynstat
#if defined ( XBNBT_GD )
		if( m_bDynstatGenerate && !m_strDynstatFontFile.empty( ) && !m_strDynstatLinkURL.empty( ) )
			pResponse->strContent += Button_JS_Link( "link_dynstat", gmapLANG_CFG["admin_update_dynstat"], "admin_dynstat()" );
#endif

		pResponse->strContent += "</p>\n";

		//
		// kill tracker
		//  

		pResponse->strContent += "<p class=\"tracker_kill\">";
		pResponse->strContent += Button_JS_Link( "confirm_kill", gmapLANG_CFG["admin_kill_tracker"], "kill_confirm()" );
		pResponse->strContent += "</p>\n<p class=\"tracker_kill_warning\">" + gmapLANG_CFG["admin_kill_warning"] + "</p>\n";

		// Output common HTML tail
		HTML_Common_End( pRequest, pResponse, btv, NOT_INDEX, string( CSS_ADMIN ) );
	}
	else
	{
		// not authorized

		// Output common HTML head
		HTML_Common_Begin(  pRequest, pResponse, gmapLANG_CFG["admin_page"], string( CSS_ADMIN ), string( ), NOT_INDEX, CODE_401 );

		// Output common HTML tail
		HTML_Common_End( pRequest, pResponse, btv, NOT_INDEX, string( CSS_ADMIN ) );
	}
}

void CTracker :: serverResponseLanguage( struct request_t *pRequest, struct response_t *pResponse )
{
	// Set the start time
	const struct bnbttv btv( UTIL_CurrentTime( ) );

	// Verify that the IP is permitted to access the tracker
	if( m_ucIPBanMode != 0)
		if( IsIPBanned( pRequest, pResponse, btv, gmapLANG_CFG["admin_page"], string( CSS_ADMIN ), NOT_INDEX ) )
			return;

	if( pRequest->user.ucAccess & ACCESS_ADMIN )
	{
		// submit?
		if( pRequest->mapParams["ap_submit_field"] == "1" )
		{
			if( !pRequest->mapParams["map"].empty( ) )
				LANG_CFG_SetString( pRequest->mapParams["map"], pRequest->mapParams["msg"] );
			
			LANG_CFG_Close( LANG_CFG_FILE );

			return JS_ReturnToPage( pRequest, pResponse, LANGUAGE_HTML );
		}
		
		// default?
		if( pRequest->mapParams["ap_default_field"] == "1" )
		{
			if( !pRequest->mapParams["map"].empty( ) )
				LANG_CFG_Delete( pRequest->mapParams["map"] );
			
			LANG_CFG_Close( LANG_CFG_FILE );
			gmapLANG_CFG.clear( );

			LANG_CFG_Init( LANG_CFG_FILE );

			return JS_ReturnToPage( pRequest, pResponse, LANGUAGE_HTML );
		}

		//
		// display the language page
		//

		// Output common HTML head
		HTML_Common_Begin( pRequest, pResponse, gmapLANG_CFG["admin_page"], string( CSS_ADMIN ), string( ), NOT_INDEX, CODE_200 );

		// Javascript
		pResponse->strContent += "<script type=\"text/javascript\">\n";
		pResponse->strContent += "<!--\n";
		pResponse->strContent += "function submit_language_confirm( ID, MAP )\n";
		pResponse->strContent += "{\n";
		pResponse->strContent += "var MSG=document.languageedit.elements[ID].value\n";
		pResponse->strContent += "var name=confirm(\"" + gmapLANG_CFG["Submit"] + "\\n\" + MAP + \"\\n\" + MSG)\n";
		pResponse->strContent += "if (name==true)\n";
		pResponse->strContent += "{\n";
		pResponse->strContent += "window.location=\"" + RESPONSE_STR_LANGUAGE_HTML + "?ap_submit_field=1&map=\" + MAP + \"&msg=\" + escape(MSG)\n";;
		pResponse->strContent += "}\n";
		pResponse->strContent += "}\n\n";
		pResponse->strContent += "function default_language_confirm( MAP )\n";
		pResponse->strContent += "{\n";
		pResponse->strContent += "var name=confirm(\"Default\\n\" + MAP)\n";
		pResponse->strContent += "if (name==true)\n";
		pResponse->strContent += "{\n";
		pResponse->strContent += "window.location=\"" + RESPONSE_STR_LANGUAGE_HTML + "?ap_default_field=1&map=\" + MAP\n";;
		pResponse->strContent += "}\n";
		pResponse->strContent += "}\n\n";
		pResponse->strContent += "//-->\n";
		pResponse->strContent += "</script>\n\n";

		// table
		pResponse->strContent += "<div class=\"language_table\">\n";
		pResponse->strContent += "<form method=\"get\" name=\"languageedit\" onSubmit=\"return confirm(\'" + gmapLANG_CFG["Submit"] + "\')\" action=\"" + RESPONSE_STR_LANGUAGE_HTML + "\">\n";
		pResponse->strContent += "<table summary=\"xadminlanguage\">\n";

		pResponse->strContent += "<tr>\n";
		pResponse->strContent += "<th colspan=3>Admin Language</th>";
		pResponse->strContent += "</tr>\n";

		pResponse->strContent += "<tr>\n";
		pResponse->strContent += "<th colspan=1>Message ID</th>";
		pResponse->strContent += "<th colspan=1>Message</th>";
		pResponse->strContent += "<th colspan=1>Action</th>";
		pResponse->strContent += "</tr>\n";

		unsigned int uiRow = 0;
		unsigned int uiElement = 0;

		for( map<string, string>::iterator pos = gmapLANG_CFG.begin(); pos != gmapLANG_CFG.end(); pos++)
		{
			if( ( ++uiRow ) % 2 )
				pResponse->strContent += "<tr class=\"even\">\n";
			else
				pResponse->strContent += "<tr class=\"odd\">\n";

			pResponse->strContent += "<th>" + (*pos).first + "</th>\n";
			pResponse->strContent += "<td><span style=\"display:none\"><label for=\"id_" + (*pos).first + "\">" + (*pos).first + "</label></span><input name=\"name_" + (*pos).first + "\" id=\"id_" + (*pos).first + "\" type=text size=140 value=\"" + UTIL_RemoveHTML( (*pos).second ) + "\"></td>\n";

			pResponse->strContent += "<td>";
			pResponse->strContent += Button_JS_Link( (*pos).first + "_submit", gmapLANG_CFG["Submit"], "submit_language_confirm( " + CAtomInt( uiElement ).toString( ) + ", \'" + (*pos).first + "\' )" );
			pResponse->strContent += Button_Reset( (*pos).first + "_reset", gmapLANG_CFG["reset"] );
			pResponse->strContent += Button_JS_Link( (*pos).first + "_default", "Default", "default_language_confirm( \'" + (*pos).first + "\' )" );
			pResponse->strContent += "</td>\n";

			pResponse->strContent += "</tr>\n";

			uiElement += 4;
		}

		pResponse->strContent += "</table>\n</form>\n</div>\n\n";

		// Output common HTML tail
		HTML_Common_End( pRequest, pResponse, btv, NOT_INDEX, string( CSS_ADMIN ) );
	}
	else
	{
		// not authorized

		// Output common HTML head
		HTML_Common_Begin(  pRequest, pResponse, gmapLANG_CFG["admin_page"], string( CSS_ADMIN ), string( ), NOT_INDEX, CODE_401 );

		// Output common HTML tail
		HTML_Common_End( pRequest, pResponse, btv, NOT_INDEX, string( CSS_ADMIN ) );
	}
}

#if defined ( XBNBT_GD )
// XBNBT Reset Dynstats Statistics
void CTracker :: resetXStatsDynstat( )
{
	gtXStats.dynstat.iFrozen = 0;
	gtXStats.dynstat.iProcessed = 0;
	gtXStats.dynstat.iRun = 0;
	gtXStats.dynstat.iTotalProcessed = 0;
	gtXStats.dynstat.sElapsedTime = string( );
	gtXStats.dynstat.sLastRunTime = string( );
	gtXStats.dynstat.sLastReset = UTIL_Date( );
}
#endif

// XBNBT Reset Announce Statistics
void CTracker :: resetXStatsAnnounce( )
{
	gtXStats.announce.iAnnounce = 0;
	gtXStats.announce.iAnnMissing = 0;
	gtXStats.announce.iAnnNotAuth = 0;
	gtXStats.announce.iCompact = 0;
	gtXStats.announce.iNopeerid = 0;
	gtXStats.announce.iRegular = 0;
	gtXStats.announce.iIPBanned = 0;
	gtXStats.announce.iIPNotCleared = 0;
	//gtXStats.announce.iIPNATBlocked = 0;
	//gtXStats.announce.iIPLocalOnly = 0;
	gtXStats.announce.iPeerSpoofRestrict = 0;
	gtXStats.announce.iClientBanned = 0;
	gtXStats.announce.iInvalidEvent = 0;
	gtXStats.announce.iPortMissing = 0;
	gtXStats.announce.iUploadedMissing = 0;
	gtXStats.announce.iDownloadedMissing = 0;
	gtXStats.announce.iLeftMissing = 0;
	gtXStats.announce.iPeerIDLength = 0;
	gtXStats.announce.iPortBlacklisted = 0;
	gtXStats.announce.iDownloadedInvalid = 0;
	gtXStats.announce.iUploadedInvalid = 0;
	gtXStats.announce.iNotAuthorized = 0;
	gtXStats.announce.sLastReset = UTIL_Date( );
}

// XBNBT Reset Scrape Statistics
void CTracker :: resetXStatsScrape( )
{
	gtXStats.scrape.iScrape = 0;
	gtXStats.scrape.iDisallowed = 0;
	gtXStats.scrape.iSingle = 0;
	gtXStats.scrape.iMultiple = 0;
	gtXStats.scrape.iFull = 0;
	gtXStats.scrape.iNotAuthorized = 0;
	gtXStats.scrape.sLastReset = UTIL_Date( );
}

// XBNBT Reset File Statistics
void CTracker :: resetXStatsFile( )
{
	gtXStats.file.iBarFill = 0;
	gtXStats.file.iBarTrans = 0;
	gtXStats.file.iCSS = 0;
	gtXStats.file.iFavicon = 0;
	gtXStats.file.uiFiles = 0;
	gtXStats.file.iRobots = 0;
	gtXStats.file.iRSS = 0;
	gtXStats.file.iRSSXSL = 0;
	gtXStats.file.iXML = 0;
	gtXStats.file.iTorrent = 0;
	gtXStats.file.sLastReset = UTIL_Date( );
}

// XBNBT Reset Page Statistics
void CTracker :: resetXStatsPage( )
{
	gtXStats.page.iAdmin = 0;
	gtXStats.page.uiComments = 0;
	gtXStats.page.iError404 = 0;
	gtXStats.page.iIndex = 0;
	gtXStats.page.iInfo = 0;
	gtXStats.page.iLogin = 0;
	gtXStats.page.iSignup = 0;
	gtXStats.page.iStats = 0;
	gtXStats.page.iTorrents = 0;
	gtXStats.page.iUpload = 0;
	gtXStats.page.iUsers = 0;
	gtXStats.page.iTags = 0;
	gtXStats.page.iLanguage = 0;
	gtXStats.page.iXStats = 0;
	gtXStats.page.sLastReset = UTIL_Date( );
}

// XBNBT Reset TCP Statistics
void CTracker :: resetXStatsTCP( )
{
	gtXStats.tcp.iRecv = 0;
	gtXStats.tcp.iSend = 0;
	gtXStats.tcp.iRecvHub = 0;
	gtXStats.tcp.iSendHub = 0;
	gtXStats.tcp.iRecvLink = 0;
	gtXStats.tcp.iSendLink = 0;
	gtXStats.tcp.sLastReset = UTIL_Date( );
}

// XBNBT Reset Peer Statistics
void CTracker :: resetXStatsPeers( )
{
	gtXStats.peer.iGreatest = 0;
	gtXStats.peer.iGreatestSeeds = 0;
	gtXStats.peer.iGreatestLeechers = 0;
	gtXStats.peer.iGreatestUnique = 0;
	gtXStats.peer.sLastReset = UTIL_Date( );
}

// XBNBT Reset All Statistics
void CTracker :: resetXStatsAll( )
{
#if defined ( XBNBT_GD )
	resetXStatsDynstat( );
#endif
	resetXStatsAnnounce( );
	resetXStatsScrape( );
	resetXStatsFile( );
	resetXStatsPage( );
	resetXStatsTCP( );
	resetXStatsPeers( );
}

// XBNBT Stats EXport Stats from struct to dictionary and save to file
void CTracker :: saveXStats( )
{
	if( m_pXStats )
	{
		if( !m_pXStats->getItem( "announce" ) )
			m_pXStats->setItem( "announce", new CAtomDicti( ) );

		CAtom *pXAnnounce = m_pXStats->getItem( "announce" );

		if( pXAnnounce && pXAnnounce->isDicti( ) )
		{
			( (CAtomDicti *)pXAnnounce )->setItem( "iAnnounce", new CAtomLong( gtXStats.announce.iAnnounce ) );
			( (CAtomDicti *)pXAnnounce )->setItem( "iAnnMissing", new CAtomLong( gtXStats.announce.iAnnMissing ) );
			( (CAtomDicti *)pXAnnounce )->setItem( "iAnnNotAuth", new CAtomLong( gtXStats.announce.iAnnNotAuth ) );
			( (CAtomDicti *)pXAnnounce )->setItem( "iClientBanned", new CAtomLong( gtXStats.announce.iClientBanned ) );
			( (CAtomDicti *)pXAnnounce )->setItem( "iCompact", new CAtomLong( gtXStats.announce.iCompact ) );
			( (CAtomDicti *)pXAnnounce )->setItem( "iDownloadedInvalid", new CAtomLong( gtXStats.announce.iDownloadedInvalid) );
			( (CAtomDicti *)pXAnnounce )->setItem( "iDownloadedMissing", new CAtomLong( gtXStats.announce.iDownloadedMissing ) );
			( (CAtomDicti *)pXAnnounce )->setItem( "iInvalidEvent", new CAtomLong( gtXStats.announce.iInvalidEvent ) );
			( (CAtomDicti *)pXAnnounce )->setItem( "iIPBanned", new CAtomLong( gtXStats.announce.iIPBanned ) );
			( (CAtomDicti *)pXAnnounce )->setItem( "iIPNotCleared", new CAtomLong( gtXStats.announce.iIPNotCleared ) );
			( (CAtomDicti *)pXAnnounce )->setItem( "iLeftMissing", new CAtomLong( gtXStats.announce.iLeftMissing ) );
			( (CAtomDicti *)pXAnnounce )->setItem( "iNopeerid", new CAtomLong( gtXStats.announce.iNopeerid ) );
			( (CAtomDicti *)pXAnnounce )->setItem( "iPeerIDLength", new CAtomLong( gtXStats.announce.iPeerIDLength ) );
			( (CAtomDicti *)pXAnnounce )->setItem( "iPeerSpoofRestrict", new CAtomLong( gtXStats.announce.iPeerSpoofRestrict ) );
			( (CAtomDicti *)pXAnnounce )->setItem( "iPortBlacklisted", new CAtomLong( gtXStats.announce.iPortBlacklisted ) );
			( (CAtomDicti *)pXAnnounce )->setItem( "iPortMissing", new CAtomLong( gtXStats.announce.iPortMissing ) );
			( (CAtomDicti *)pXAnnounce )->setItem( "iRegular", new CAtomLong( gtXStats.announce.iRegular ) );
			( (CAtomDicti *)pXAnnounce )->setItem( "iUploadedInvalid", new CAtomLong( gtXStats.announce.iUploadedInvalid) );
			( (CAtomDicti *)pXAnnounce )->setItem( "iUploadedMissing", new CAtomLong( gtXStats.announce.iUploadedMissing ) );
			( (CAtomDicti *)pXAnnounce )->setItem( "iNotAuthorized", new CAtomLong( gtXStats.announce.iNotAuthorized ) );
			( (CAtomDicti *)pXAnnounce )->setItem( "sLastReset", new CAtomString( gtXStats.announce.sLastReset ) );
		}

#if defined ( XBNBT_GD )
		if( !m_pXStats->getItem( "dynstat" ) )
			m_pXStats->setItem( "dynstat", new CAtomDicti( ) );

		CAtom *pXDynstat = m_pXStats->getItem( "dynstat" );

		if( pXDynstat && pXDynstat->isDicti( ) )
		{
			( (CAtomDicti *)pXDynstat )->setItem( "iFrozen", new CAtomLong( gtXStats.dynstat.iFrozen ) );
			( (CAtomDicti *)pXDynstat )->setItem( "iProcessed", new CAtomLong( gtXStats.dynstat.iProcessed ) );
			( (CAtomDicti *)pXDynstat )->setItem( "iRun", new CAtomLong( gtXStats.dynstat.iRun ) );
			( (CAtomDicti *)pXDynstat )->setItem( "iTotalProcessed", new CAtomLong( gtXStats.dynstat.iTotalProcessed ) );
			( (CAtomDicti *)pXDynstat )->setItem( "sLastRunTime", new CAtomString( gtXStats.dynstat.sLastRunTime ) );
			( (CAtomDicti *)pXDynstat )->setItem( "sElapsedTime", new CAtomString( gtXStats.dynstat.sElapsedTime ) );
			( (CAtomDicti *)pXDynstat )->setItem( "sLastReset", new CAtomString( gtXStats.dynstat.sLastReset ) );
		}
#endif

		if( !m_pXStats->getItem( "file" ) )
			m_pXStats->setItem( "file", new CAtomDicti( ) );

		CAtom *pXFile = m_pXStats->getItem( "file" );

		if( pXFile && pXFile->isDicti( ) )
		{
			( (CAtomDicti *)pXFile )->setItem( "iBarFill", new CAtomLong( gtXStats.file.iBarFill ) );
			( (CAtomDicti *)pXFile )->setItem( "iBarTrans", new CAtomLong( gtXStats.file.iBarTrans ) );
			( (CAtomDicti *)pXFile )->setItem( "iCSS", new CAtomLong( gtXStats.file.iCSS ) );
			( (CAtomDicti *)pXFile )->setItem( "iFavicon", new CAtomLong( gtXStats.file.iFavicon ) );
			( (CAtomDicti *)pXFile )->setItem( "uiFiles", new CAtomLong( gtXStats.file.uiFiles ) );
			( (CAtomDicti *)pXFile )->setItem( "iRobots", new CAtomLong( gtXStats.file.iRobots ) );
			( (CAtomDicti *)pXFile )->setItem( "iRSS", new CAtomLong( gtXStats.file.iRSS ) );
			( (CAtomDicti *)pXFile )->setItem( "iRSSXSL", new CAtomLong( gtXStats.file.iRSSXSL ) );
			( (CAtomDicti *)pXFile )->setItem( "iXML", new CAtomLong( gtXStats.file.iXML ) );
			( (CAtomDicti *)pXFile )->setItem( "iXML", new CAtomLong( gtXStats.file.iTorrent ) );
			( (CAtomDicti *)pXFile )->setItem( "sLastReset", new CAtomString( gtXStats.file.sLastReset ) );
		}

		if( !m_pXStats->getItem( "page" ) )
			m_pXStats->setItem( "page", new CAtomDicti( ) );

		CAtom *pXPage = m_pXStats->getItem( "page" );

		if( pXPage && pXPage->isDicti( ) )
		{
			( (CAtomDicti *)pXPage )->setItem( "iAdmin", new CAtomLong( gtXStats.page.iAdmin ) );
			( (CAtomDicti *)pXPage )->setItem( "uiComments", new CAtomLong( gtXStats.page.uiComments ) );
			( (CAtomDicti *)pXPage )->setItem( "iError404", new CAtomLong( gtXStats.page.iError404 ) );
			( (CAtomDicti *)pXPage )->setItem( "iIndex", new CAtomLong( gtXStats.page.iIndex ) );
			( (CAtomDicti *)pXPage )->setItem( "iInfo", new CAtomLong( gtXStats.page.iInfo ) );
			( (CAtomDicti *)pXPage )->setItem( "iLogin", new CAtomLong( gtXStats.page.iLogin ) );
			( (CAtomDicti *)pXPage )->setItem( "iSignup", new CAtomLong( gtXStats.page.iSignup ) );
			( (CAtomDicti *)pXPage )->setItem( "iStats", new CAtomLong( gtXStats.page.iStats ) );
			( (CAtomDicti *)pXPage )->setItem( "iTorrents", new CAtomLong( gtXStats.page.iTorrents) );
			( (CAtomDicti *)pXPage )->setItem( "iUpload", new CAtomLong( gtXStats.page.iUpload ) );
			( (CAtomDicti *)pXPage )->setItem( "iUsers", new CAtomLong( gtXStats.page.iUsers ) );
			( (CAtomDicti *)pXPage )->setItem( "iTags", new CAtomLong( gtXStats.page.iTags ) );
			( (CAtomDicti *)pXPage )->setItem( "iLanguage", new CAtomLong( gtXStats.page.iLanguage ) );
			( (CAtomDicti *)pXPage )->setItem( "iXStats", new CAtomLong( gtXStats.page.iXStats ) );
			( (CAtomDicti *)pXPage )->setItem( "sLastReset", new CAtomString( gtXStats.page.sLastReset ) );
		}

		if( !m_pXStats->getItem( "scrape" ) )
			m_pXStats->setItem( "scrape", new CAtomDicti( ) );

		CAtom *pXScrape = m_pXStats->getItem( "scrape" );

		if( pXScrape && pXScrape->isDicti( ) )
		{
			( (CAtomDicti *)pXScrape )->setItem( "iScrape", new CAtomLong( gtXStats.scrape.iScrape ) );
			( (CAtomDicti *)pXScrape )->setItem( "iDisallowed", new CAtomLong( gtXStats.scrape.iDisallowed ) );
			( (CAtomDicti *)pXScrape )->setItem( "iSingle", new CAtomLong( gtXStats.scrape.iSingle ) );
			( (CAtomDicti *)pXScrape )->setItem( "iMultiple", new CAtomLong( gtXStats.scrape.iMultiple ) );
			( (CAtomDicti *)pXScrape )->setItem( "iFull", new CAtomLong( gtXStats.scrape.iFull ) );
			( (CAtomDicti *)pXScrape )->setItem( "sLastReset", new CAtomString( gtXStats.scrape.sLastReset ) );
		}

		if( !m_pXStats->getItem( "tcp" ) )
			m_pXStats->setItem( "tcp", new CAtomDicti( ) );

		CAtom *pXTCP = m_pXStats->getItem( "tcp" );

		if( pXTCP && pXTCP->isDicti( ) )
		{
			( (CAtomDicti *)pXTCP )->setItem( "iRecv", new CAtomLong( gtXStats.tcp.iRecv ) );
			( (CAtomDicti *)pXTCP )->setItem( "iSend", new CAtomLong( gtXStats.tcp.iSend ) );
			( (CAtomDicti *)pXTCP )->setItem( "iRecvHub", new CAtomLong( gtXStats.tcp.iRecvHub ) );
			( (CAtomDicti *)pXTCP )->setItem( "iSendHub", new CAtomLong( gtXStats.tcp.iSendHub ) );
			( (CAtomDicti *)pXTCP )->setItem( "iRecvLink", new CAtomLong( gtXStats.tcp.iRecvLink ) );
			( (CAtomDicti *)pXTCP )->setItem( "iSendLink", new CAtomLong( gtXStats.tcp.iSendLink ) );
			( (CAtomDicti *)pXTCP )->setItem( "sLastReset", new CAtomString( gtXStats.tcp.sLastReset ) );
		}

		if( !m_pXStats->getItem( "peer" ) )
			m_pXStats->setItem( "peer", new CAtomDicti( ) );

		CAtom *pXPeer = m_pXStats->getItem( "peer" );

		if( pXPeer && pXPeer->isDicti( ) )
		{
			( (CAtomDicti *)pXPeer )->setItem( "iGreatest", new CAtomLong( gtXStats.peer.iGreatest ) );
			( (CAtomDicti *)pXPeer )->setItem( "iGreatestSeeds", new CAtomLong( gtXStats.peer.iGreatestSeeds ) );
			( (CAtomDicti *)pXPeer )->setItem( "iGreatestLeechers", new CAtomLong( gtXStats.peer.iGreatestLeechers ) );
			( (CAtomDicti *)pXPeer )->setItem( "iGreatestUnique", new CAtomLong( gtXStats.peer.iGreatestUnique ) );
			( (CAtomDicti *)pXPeer )->setItem( "sLastReset", new CAtomString( gtXStats.peer.sLastReset ) );
		}
	}

	const string cstrData( Encode( m_pXStats ) );

	FILE *pFile = FILE_ERROR;

	pFile = fopen( statsdump.strName.c_str( ), "wb" );

	if( pFile == FILE_ERROR )
	{
		UTIL_LogPrint( ( gmapLANG_CFG["unable_to_write_file"] + "\n" ).c_str( ), statsdump.strName.c_str( ) );

		return;
	}

	fwrite( cstrData.c_str( ), sizeof( char ), cstrData.size( ), pFile );
	fclose( pFile );
}

#if defined ( XBNBT_GD )
// XBNBT XStats table for Dynstat
void CTracker :: tableXStatsDynstat( struct request_t *pRequest, struct response_t *pResponse )
{
	pResponse->strContent += "<div class=\"admin_table\">\n";
	pResponse->strContent += "<table summary=\"xdsstats\">\n";

	if( pRequest->user.ucAccess & ACCESS_ADMIN )
	{
		pResponse->strContent += "<tr><th colspan=6>" + gmapLANG_CFG["stats_dynstat"] + "/th>\n";
		pResponse->strContent += "<th colspan=1>";
		pResponse->strContent += Button_JS_Link( "confirm_dynstat", gmapLANG_CFG["reset"], "dynstat_confirm()" );
		pResponse->strContent += "</th></tr>\n";
	}
	else
		pResponse->strContent += "<tr><th colspan=7>" + gmapLANG_CFG["stats_dynstat"] + "</th></tr>\n";

	pResponse->strContent += "<tr>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_last_reset"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_run_times"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_last_run"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_last_elapsed"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_last_processed"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_torrent_frozen"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_processed_start"] + "</th>\n";
	pResponse->strContent += "</tr>\n";

	pResponse->strContent += "<tr>\n";
	pResponse->strContent += "<td class=\"even\">" + gtXStats.dynstat.sLastReset + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( gtXStats.dynstat.iRun ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"even\">" + gtXStats.dynstat.sLastRunTime + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + gtXStats.dynstat.sElapsedTime + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( gtXStats.dynstat.iProcessed ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"even\">" + CAtomLong( gtXStats.dynstat.iFrozen ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( gtXStats.dynstat.iTotalProcessed ).toString( ) + "</td>\n";
	pResponse->strContent += "</tr>\n";

	pResponse->strContent += "</table>\n";
	pResponse->strContent += "</div>\n\n";
}
#endif

// XBNBT XStats table for Announce
void CTracker :: tableXStatsAnnounce( struct request_t *pRequest, struct response_t *pResponse )
{
	pResponse->strContent += "<div class=\"admin_table\">\n";
	pResponse->strContent += "<table summary=\"xannstats\">\n";

	if( pRequest->user.ucAccess & ACCESS_ADMIN )
	{
		pResponse->strContent += "<tr><th colspan=6>" + gmapLANG_CFG["stats_announce"] + "</th>\n";
		pResponse->strContent += "<th colspan=1>";
		pResponse->strContent += Button_JS_Link( "confirm_announce", gmapLANG_CFG["reset"], "announce_confirm()" );
		pResponse->strContent += "</th></tr>\n";
	}
	else
		pResponse->strContent += "<tr><th colspan=7>" + gmapLANG_CFG["stats_announce"] + "</th></tr>\n";

	pResponse->strContent += "<tr>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_last_reset"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_announces"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_announces_missing"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_announces_notauth"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_compact"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_no_peer_id"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_regular"] + "</th>\n";
	pResponse->strContent += "</tr>\n";

	pResponse->strContent += "<tr>\n";
	pResponse->strContent += "<td class=\"even\">" + gtXStats.announce.sLastReset + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( gtXStats.announce.iAnnounce ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( gtXStats.announce.iAnnMissing ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( gtXStats.announce.iAnnNotAuth ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"even\">" + CAtomLong( gtXStats.announce.iCompact ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( gtXStats.announce.iNopeerid ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"even\">" + CAtomLong( gtXStats.announce.iRegular ).toString( ) + "</td>\n";
	pResponse->strContent += "</tr>\n";

	pResponse->strContent += "<tr>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_ip_banned"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_ip_not_cleared"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_peer_spoof_blocked"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_client_banned"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_invalid_event"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_port_missing"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_uploaded_missing"] + "</th>\n";
	pResponse->strContent += "</tr>\n";

	pResponse->strContent += "<tr>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( gtXStats.announce.iIPBanned ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"even\">" + CAtomLong( gtXStats.announce.iIPNotCleared ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( gtXStats.announce.iPeerSpoofRestrict ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( gtXStats.announce.iClientBanned ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"even\">" + CAtomLong( gtXStats.announce.iInvalidEvent ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( gtXStats.announce.iPortMissing ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"even\">" + CAtomLong( gtXStats.announce.iUploadedMissing ).toString( ) + "</td>\n";
	pResponse->strContent += "</tr>\n";

	pResponse->strContent += "<tr>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_downloaded_missing"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_left_missing"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_peer_id_length"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_port_blacklisted"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_download_invalid"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_upload_invalid"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["server_response_401"] + "</th>\n";
	pResponse->strContent += "</tr>\n";

	pResponse->strContent += "<tr>\n";
	pResponse->strContent += "<td class=\"even\">" + CAtomLong( gtXStats.announce.iDownloadedMissing ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( gtXStats.announce.iLeftMissing ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"even\">" + CAtomLong( gtXStats.announce.iPeerIDLength ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( gtXStats.announce.iPortBlacklisted ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"even\">" + CAtomLong( gtXStats.announce.iDownloadedInvalid ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( gtXStats.announce.iUploadedInvalid ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"even\">" + CAtomLong( gtXStats.announce.iNotAuthorized ).toString( ) + "</td>\n";
	pResponse->strContent += "</tr>\n";

	pResponse->strContent += "</table>\n";
	pResponse->strContent += "</div>\n\n";
}

// XBNBT XStats table for Scrape
void CTracker :: tableXStatsScrape( struct request_t *pRequest, struct response_t *pResponse )
{
	pResponse->strContent += "<div class=\"admin_table\">\n";
	pResponse->strContent += "<table summary=\"xscrstats\">\n";

	if( pRequest->user.ucAccess & ACCESS_ADMIN )
	{
		pResponse->strContent += "<tr>\n";
		pResponse->strContent += "<th colspan=5>";
		pResponse->strContent += gmapLANG_CFG["stats_scrape"] + "</th>\n";
		pResponse->strContent += "<th colspan=1>";
		pResponse->strContent += Button_JS_Link( "confirm_scrape", gmapLANG_CFG["reset"], "scrape_confirm()" );
		pResponse->strContent += "</th>\n</tr>\n";
	}
	else
	{
		pResponse->strContent += "<tr>\n";
		pResponse->strContent += "<th colspan=6>";
		pResponse->strContent += gmapLANG_CFG["stats_scrape"] + "</th>\n</tr>\n";
	}

	pResponse->strContent += "<tr>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_last_reset"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_scrapes"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_single"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_multiple"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_full"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["server_response_401"] + "</th>\n";
	pResponse->strContent += "</tr>\n";

	pResponse->strContent += "<tr>\n";
	pResponse->strContent += "<td class=\"even\">" + gtXStats.scrape.sLastReset + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( gtXStats.scrape.iScrape ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"even\">" + CAtomLong( gtXStats.scrape.iSingle ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( gtXStats.scrape.iMultiple ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"even\">" + CAtomLong( gtXStats.scrape.iFull ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( gtXStats.scrape.iDisallowed ).toString( ) + "</td>\n";
	pResponse->strContent += "</tr>\n";

	pResponse->strContent += "</table>\n";
	pResponse->strContent += "</div>\n\n";
}

// XBNBT XStats table for File
void CTracker :: tableXStatsFile( struct request_t *pRequest, struct response_t *pResponse )
{
	pResponse->strContent += "<div class=\"admin_table\">\n";
	pResponse->strContent += "<table summary=\"xfilestats\">\n";

	if( pRequest->user.ucAccess & ACCESS_ADMIN )
	{
		pResponse->strContent += "<tr><th colspan=10>" + gmapLANG_CFG["stats_file"] + "</th>\n";
		pResponse->strContent += "<th colspan=1>";
		pResponse->strContent += Button_JS_Link( "confirm_file", gmapLANG_CFG["reset"], "file_confirm()" );
		pResponse->strContent += "</th></tr>\n";
	}
	else
	{
		pResponse->strContent += "<tr><th colspan=11>" + gmapLANG_CFG["stats_file"] + "</th></tr>\n";
	}

	pResponse->strContent += "<tr>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_last_reset"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_favicon"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_ibfill"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_ibtrans"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_css"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_rss"] + "</th>\n";
	pResponse->strContent += "<th>RSS.XSL</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_xml"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_robots"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_torrents"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_other"] + "</th>\n";
	pResponse->strContent += "</tr>\n";
	pResponse->strContent += "<tr>\n";
	pResponse->strContent += "<td class=\"even\">" + gtXStats.file.sLastReset + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( gtXStats.file.iFavicon ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"even\">" + CAtomLong( gtXStats.file.iBarFill ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( gtXStats.file.iBarTrans ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"even\">" + CAtomLong( gtXStats.file.iCSS ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( gtXStats.file.iRSS ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"even\">" + CAtomLong( gtXStats.file.iRSSXSL ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( gtXStats.file.iXML ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"even\">" + CAtomLong( gtXStats.file.iRobots ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( gtXStats.file.iTorrent ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"even\">" + CAtomLong( gtXStats.file.uiFiles ).toString( ) + "</td>\n";
	pResponse->strContent += "</tr>\n";

	pResponse->strContent += "</table>\n";
	pResponse->strContent += "</div>\n\n";
}

// XBNBT XStats table for Page
void CTracker :: tableXStatsPage( struct request_t *pRequest, struct response_t *pResponse )
{
	pResponse->strContent += "<div class=\"admin_table\">\n";
	pResponse->strContent += "<table summary=\"xpagestats\">\n";

	if( pRequest->user.ucAccess & ACCESS_ADMIN )
	{
		pResponse->strContent += "<tr><th colspan=7>XBNBT HTML Page Stats</th>\n";
		pResponse->strContent += "<th colspan=1>";
		pResponse->strContent += Button_JS_Link( "confirm_page", gmapLANG_CFG["reset"], "page_confirm()" );
		pResponse->strContent += "</th></tr>\n";
	}
	else
	{
		pResponse->strContent += "<tr><th colspan=8>XBNBT HTML Page Stats</th></tr>\n";
	}

	pResponse->strContent += "<tr>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_last_reset"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_index"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_stats"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_comments"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_info"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_login"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_signup"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_upload"] + "</th>\n";
	pResponse->strContent += "</tr>\n";

	pResponse->strContent += "<tr>\n";
	pResponse->strContent += "<td class=\"even\">" + gtXStats.page.sLastReset + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( gtXStats.page.iIndex ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"even\">" + CAtomLong( gtXStats.page.iStats ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( gtXStats.page.uiComments ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"even\">" + CAtomLong( gtXStats.page.iInfo ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( gtXStats.page.iLogin ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"even\">" + CAtomLong( gtXStats.page.iSignup ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( gtXStats.page.iUpload ).toString( ) + "</td>\n";
	pResponse->strContent += "</tr>\n";

	pResponse->strContent += "<tr>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_torrents"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_admin"] + "</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_users"] + "</th>\n";
	pResponse->strContent += "<th>Tags</th>\n";
	pResponse->strContent += "<th>Language</th>\n";
	pResponse->strContent += "<th>XStats</th>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_error404"] + "</th>\n";
	pResponse->strContent += "<th></th>\n";
	pResponse->strContent += "</tr>\n";

	pResponse->strContent += "<tr>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( gtXStats.page.iTorrents ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"even\">" + CAtomLong( gtXStats.page.iAdmin ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( gtXStats.page.iUsers ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"even\">" + CAtomLong( gtXStats.page.iTags ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( gtXStats.page.iLanguage ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"even\">" + CAtomLong( gtXStats.page.iXStats ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( gtXStats.page.iError404 ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"even\"></td>\n";
	pResponse->strContent += "</tr>\n";

	pResponse->strContent += "</table>\n";
	pResponse->strContent += "</div>\n\n";
}

// XBNBT XStats table for TCP
void CTracker :: tableXStatsTCP( struct request_t *pRequest, struct response_t *pResponse )
{
	pResponse->strContent += "<div class=\"admin_table\">\n";
	pResponse->strContent += "<table summary=\"xtcpstats\">\n";

	if( pRequest->user.ucAccess & ACCESS_ADMIN )
	{
		pResponse->strContent += "<tr>\n";
	
		pResponse->strContent += "<th colspan=8>TCP Stats</th>\n";

		pResponse->strContent += "<th colspan=1>";
		pResponse->strContent += Button_JS_Link( "confirm_tcp", gmapLANG_CFG["reset"], "tcp_confirm()" );
		pResponse->strContent += "</th>\n</tr>\n";
	}
	else
	{
		pResponse->strContent += "<tr>\n";

		pResponse->strContent += "<th colspan=9>TCP Stats</th>\n</tr>\n";
	}

	pResponse->strContent += "<tr>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_last_reset"] + "</th>\n";
	pResponse->strContent += "<th>Total Received</th>\n";
	pResponse->strContent += "<th>Total Sent</th>\n";
	pResponse->strContent += "<th>Received Client</th>\n";
	pResponse->strContent += "<th>Sent Client</th>\n";
	pResponse->strContent += "<th>Received Hub</th>\n";
	pResponse->strContent += "<th>Sent Hub</th>\n";
	pResponse->strContent += "<th>Received Link</th>\n";
	pResponse->strContent += "<th>Sent Link</th>\n";
	pResponse->strContent += "</tr>\n";

	pResponse->strContent += "<tr>\n";
	pResponse->strContent += "<td class=\"even\">" + gtXStats.tcp.sLastReset + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + UTIL_BytesToString( gtXStats.tcp.iRecv + gtXStats.tcp.iRecvHub + gtXStats.tcp.iRecvLink ) + "</td>\n";
	pResponse->strContent += "<td class=\"even\">" + UTIL_BytesToString( gtXStats.tcp.iSend + gtXStats.tcp.iSendHub + gtXStats.tcp.iSendLink ) + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + UTIL_BytesToString( gtXStats.tcp.iRecv ) + "</td>\n";
	pResponse->strContent += "<td class=\"even\">" + UTIL_BytesToString( gtXStats.tcp.iSend ) + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + UTIL_BytesToString( gtXStats.tcp.iRecvHub ) + "</td>\n";
	pResponse->strContent += "<td class=\"even\">" + UTIL_BytesToString( gtXStats.tcp.iSendHub ) + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + UTIL_BytesToString( gtXStats.tcp.iRecvLink ) + "</td>\n";
	pResponse->strContent += "<td class=\"even\">" + UTIL_BytesToString( gtXStats.tcp.iSendLink ) + "</td>\n";
	pResponse->strContent += "</tr>\n";

	pResponse->strContent += "</table>\n";
	pResponse->strContent += "</div>\n\n";
}

// XBNBT XStats table for Peers
void CTracker :: tableXStatsPeer( struct request_t *pRequest, struct response_t *pResponse )
{
	pResponse->strContent += "<div class=\"admin_table\">\n";
	pResponse->strContent += "<table summary=\"xpeerstats\">\n";

	if( pRequest->user.ucAccess & ACCESS_ADMIN )
	{
		pResponse->strContent += "<tr>\n";
	
		pResponse->strContent += "<th colspan=8>Peer Stats</th>\n";

		pResponse->strContent += "<th colspan=1>";
		pResponse->strContent += Button_JS_Link( "confirm_peer", gmapLANG_CFG["reset"], "peer_confirm()" );
		pResponse->strContent += "</th>\n</tr>\n";
	}
	else
	{
		pResponse->strContent += "<tr>\n";

		pResponse->strContent += "<th colspan=9>Peer Stats</th>\n</tr>\n";
	}

	pResponse->strContent += "<tr>\n";
	pResponse->strContent += "<th>" + gmapLANG_CFG["stats_last_reset"] + "</th>\n";
	pResponse->strContent += "<th>Current Peers</th>\n";
	pResponse->strContent += "<th>Current Seeds</th>\n";
	pResponse->strContent += "<th>Current Leechers</th>\n";
	pResponse->strContent += "<th>Current Unique</th>\n";
	pResponse->strContent += "<th>Greatest Peers</th>\n";
	pResponse->strContent += "<th>Greatest Seeders</th>\n";
	pResponse->strContent += "<th>Greatest Leechers</th>\n";
	pResponse->strContent += "<th>Greatest Unique</th>\n";
	pResponse->strContent += "</tr>\n";

	// Refresh the fast cache
	if( GetTime( ) > m_ulRefreshFastCacheNext )
	{
		// Refresh
		RefreshFastCache( );

		// Set the next refresh time
		m_ulRefreshFastCacheNext = GetTime( ) + m_uiRefreshFastCacheInterval;
	}

	if( m_bCountUniquePeers )
	{
		if( m_pIPs->getValuePtr( )->size( ) > gtXStats.peer.iGreatestUnique )
			gtXStats.peer.iGreatestUnique = m_pIPs->getValuePtr( )->size( );
	}

	pResponse->strContent += "<tr>\n";
	pResponse->strContent += "<td class=\"even\">" + gtXStats.peer.sLastReset + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( m_ulPeers ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"even\">" + CAtomLong( m_ulSeedersTotal ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( m_ulLeechersTotal ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"even\">" + CAtomLong( m_pIPs->getValuePtr( )->size( ) ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( gtXStats.peer.iGreatest ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"even\">" + CAtomLong( gtXStats.peer.iGreatestSeeds ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"odd\">" + CAtomLong( gtXStats.peer.iGreatestLeechers ).toString( ) + "</td>\n";
	pResponse->strContent += "<td class=\"even\">" + CAtomLong( gtXStats.peer.iGreatestUnique ).toString( ) + "</td>\n";
	pResponse->strContent += "</tr>\n";

	pResponse->strContent += "</table>\n";
	pResponse->strContent += "</div>\n\n";
}

void CTracker :: serverResponseXStats( struct request_t *pRequest, struct response_t *pResponse )
{
	// Set the start time
	const struct bnbttv btv( UTIL_CurrentTime( ) );

	// Verify that the IP is permitted to access the tracker
	if( m_ucIPBanMode != 0)
		if( IsIPBanned( pRequest, pResponse, btv, gmapLANG_CFG["admin_page"], string( CSS_ADMIN ), NOT_INDEX ) )
			return;

	// Does the user have admin access?
	if( pRequest->user.ucAccess & ACCESS_ADMIN )
	{

#if defined ( XBNBT_GD )
		if( pRequest->mapParams["ap_dynstat"] == "1" )
		{
			resetXStatsDynstat( );
			saveXStats( );

			return JS_ReturnToPage( pRequest, pResponse, XSTATS_HTML );
		}
#endif

		if( pRequest->mapParams["ap_announce"] == "1" )
		{
			resetXStatsAnnounce( );
			saveXStats( );

			return JS_ReturnToPage( pRequest, pResponse, XSTATS_HTML );
		}

		if( pRequest->mapParams["ap_scrape"] == "1" )
		{
			resetXStatsScrape( );
			saveXStats( );

			return JS_ReturnToPage( pRequest, pResponse, XSTATS_HTML );
		}

		if( pRequest->mapParams["ap_file"] == "1" )
		{
			resetXStatsFile( );
			saveXStats( );

			return JS_ReturnToPage( pRequest, pResponse, XSTATS_HTML );
		}

		if( pRequest->mapParams["ap_page"] == "1" )
		{
			resetXStatsPage( );
			saveXStats( );

			return JS_ReturnToPage( pRequest, pResponse, XSTATS_HTML );
		}

		if( pRequest->mapParams["ap_tcp"] == "1" )
		{
			resetXStatsTCP( );
			saveXStats( );

			return JS_ReturnToPage( pRequest, pResponse, XSTATS_HTML );
		}

		if( pRequest->mapParams["ap_peer"] == "1" )
		{
			resetXStatsPeers( );
			saveXStats( );

			return JS_ReturnToPage( pRequest, pResponse, XSTATS_HTML );
		}

		if( pRequest->mapParams["ap_all"] == "1" )
		{
			resetXStatsAll( );
			saveXStats( );

			return JS_ReturnToPage( pRequest, pResponse, XSTATS_HTML );
		}

		// display
		HTML_Common_Begin( pRequest, pResponse, gmapLANG_CFG["admin_page"], string( CSS_ADMIN ), string( ), NOT_INDEX, CODE_200 );
		
		// Javascript
        pResponse->strContent += "<script type=\"text/javascript\">\n";
		pResponse->strContent += "<!--\n";

		pResponse->strContent += "function dynstat_confirm()\n";
		pResponse->strContent += "{\n";
		pResponse->strContent += "var name=confirm(\"" + gmapLANG_CFG["js_reset_dynstat_stats_q"] + "\")\n";
		pResponse->strContent += "if (name==true)\n";
		pResponse->strContent += "{\n";
		pResponse->strContent += "window.location=\"" + RESPONSE_STR_XSTATS_HTML + "?ap_dynstat=1\"\n";
		pResponse->strContent += "}\n";
		pResponse->strContent += "}\n\n";

		pResponse->strContent += "function announce_confirm()\n";
		pResponse->strContent += "{\n";
		pResponse->strContent += "var name=confirm(\"" + gmapLANG_CFG["js_reset_announce_stats_q"] + "\")\n";
		pResponse->strContent += "if (name==true)\n";
		pResponse->strContent += "{\n";
		pResponse->strContent += "window.location=\"" + RESPONSE_STR_XSTATS_HTML + "?ap_announce=1\"\n";
		pResponse->strContent += "}\n";
		pResponse->strContent += "}\n\n";

		pResponse->strContent += "function scrape_confirm()\n";
		pResponse->strContent += "{\n";
		pResponse->strContent += "var name=confirm(\"" + gmapLANG_CFG["js_reset_scrape_stats_q"] + "\")\n";
		pResponse->strContent += "if (name==true)\n";
		pResponse->strContent += "{\n";
		pResponse->strContent += "window.location=\"" + RESPONSE_STR_XSTATS_HTML + "?ap_scrape=1\"\n";
		pResponse->strContent += "}\n";
		pResponse->strContent += "}\n\n";

		pResponse->strContent += "function file_confirm()\n";
		pResponse->strContent += "{\n";
		pResponse->strContent += "var name=confirm(\"" + gmapLANG_CFG["js_reset_file_stats_q"] + "\")\n";
		pResponse->strContent += "if (name==true)\n";
		pResponse->strContent += "{\n";
		pResponse->strContent += "window.location=\"" + RESPONSE_STR_XSTATS_HTML + "?ap_file=1\"\n";
		pResponse->strContent += "}\n";
		pResponse->strContent += "}\n\n";

		pResponse->strContent += "function page_confirm()\n";
		pResponse->strContent += "{\n";
		pResponse->strContent += "var name=confirm(\"" + gmapLANG_CFG["js_reset_page_stats_q"] + "\")\n";
		pResponse->strContent += "if (name==true)\n";
		pResponse->strContent += "{\n";
		pResponse->strContent += "window.location=\"" + RESPONSE_STR_XSTATS_HTML + "?ap_page=1\"\n";
		pResponse->strContent += "}\n";
		pResponse->strContent += "}\n\n";

		pResponse->strContent += "function tcp_confirm()\n";
		pResponse->strContent += "{\n";
		pResponse->strContent += "var name=confirm(\"Reset TCP Stats\")\n";
		pResponse->strContent += "if (name==true)\n";
		pResponse->strContent += "{\n";
		pResponse->strContent += "window.location=\"" + RESPONSE_STR_XSTATS_HTML + "?ap_tcp=1\"\n";
		pResponse->strContent += "}\n";
		pResponse->strContent += "}\n\n";

		pResponse->strContent += "function peer_confirm()\n";
		pResponse->strContent += "{\n";
		pResponse->strContent += "var name=confirm(\"Reset Peer Stats\")\n";
		pResponse->strContent += "if (name==true)\n";
		pResponse->strContent += "{\n";
		pResponse->strContent += "window.location=\"" + RESPONSE_STR_XSTATS_HTML + "?ap_peer=1\"\n";
		pResponse->strContent += "}\n";
		pResponse->strContent += "}\n\n";

		pResponse->strContent += "function all_confirm()\n";
		pResponse->strContent += "{\n";
		pResponse->strContent += "var name=confirm(\"" + gmapLANG_CFG["js_reset_all_stats_q"] + "\")\n";
		pResponse->strContent += "if (name==true)\n";
		pResponse->strContent += "{\n";
		pResponse->strContent += "window.location=\"" + RESPONSE_STR_XSTATS_HTML + "?ap_all=1\"\n";
		pResponse->strContent += "}\n";
		pResponse->strContent += "}\n\n";

		pResponse->strContent += "//-->\n";
		pResponse->strContent += "</script>\n\n";

		// Tables
#if defined ( XBNBT_GD )
		tableXStatsDynstat( pRequest, pResponse );
#endif
		tableXStatsAnnounce( pRequest, pResponse );
		tableXStatsScrape( pRequest, pResponse );
		tableXStatsFile( pRequest, pResponse );
		tableXStatsPage( pRequest, pResponse );
		tableXStatsTCP( pRequest, pResponse );
		tableXStatsPeer( pRequest, pResponse );

		// Button List
		pResponse->strContent += "<p class=\"button_list\">";

		// Reset all stats
		pResponse->strContent += Button_JS_Link( "confirm_all", gmapLANG_CFG["reset_all_stats"], "all_confirm()" );

		pResponse->strContent += "</p>\n";

		// Output common HTML ending
		HTML_Common_End( pRequest, pResponse, btv, NOT_INDEX, string( CSS_ADMIN ) );
	}
	else
	{
		// not authorized
		HTML_Common_Begin(  pRequest, pResponse, gmapLANG_CFG["admin_page"], string( CSS_ADMIN ), string( ), NOT_INDEX, CODE_401 );

		// Output common HTML ending
		HTML_Common_End( pRequest, pResponse, btv, NOT_INDEX, string( CSS_ADMIN ) );
	}
}

void CTracker :: serverResponseTags( struct request_t *pRequest, struct response_t *pResponse )
{
	// Set the start time
	const struct bnbttv btv( UTIL_CurrentTime( ) );

	// Verify that the IP is permitted to access the tracker
	if( m_ucIPBanMode != 0)
		if( IsIPBanned( pRequest, pResponse, btv, gmapLANG_CFG["admin_page"], string( CSS_ADMIN ), NOT_INDEX ) )
			return;

	if( pRequest->user.ucAccess & ACCESS_ADMIN )
	{
		const unsigned char cucSwapTag( (unsigned char)atoi( pRequest->mapParams["ap_swaptag"].c_str( ) ) );

		if( cucSwapTag != 0 )
		{
			unsigned char ucTag = 1;
			string strNameNext = string( );
			string strTagNext = string( );

			string strName = "bnbt_tag" + CAtomInt( ucTag ).toString( );
			string strTag = CFG_GetString( strName, string( ) );

			while( !strTag.empty( ) )
			{
				if( ucTag == cucSwapTag )
				{
					strNameNext = "bnbt_tag" + CAtomInt( ucTag + 1 ).toString( );
					strTagNext = CFG_GetString( strNameNext, string( ) );

					CFG_SetString( strName, strTagNext );
					CFG_SetString( strNameNext, strTag );

					ucTag++;
				}
				else
				{
					strName = "bnbt_tag" + CAtomInt( ucTag ).toString( );

					CFG_SetString( strName, strTag );
				}

				strName = "bnbt_tag" + CAtomInt( ++ucTag ).toString( );

				strTag = CFG_GetString( strName, string( ) );
			}

			CFG_Close( CFG_FILE );
			m_vecTags.clear( );
			m_vecTagsMouse.clear( );
			initTags( );

			return JS_ReturnToPage( pRequest, pResponse, TAGS_HTML );
		}

		const unsigned char cucDelTag( (unsigned char)atoi( pRequest->mapParams["ap_deltag"].c_str( ) ) );

		if( cucDelTag != 0 )
		{
			unsigned char ucTag = 1;
			unsigned char ucTagPrevious = 0;

			string strName = "bnbt_tag" + CAtomInt( ucTag ).toString( );
			string strTag = CFG_GetString( strName, string( ) );

			while( !strTag.empty( ) )
			{
				if( ucTag == cucDelTag )
					CFG_Delete( strName );

				if( ucTag > cucDelTag )
				{
					strName = "bnbt_tag" + CAtomInt( ucTagPrevious ).toString( );
					CFG_SetString( strName, strTag );
				}

				ucTagPrevious = ucTag;

				strName = "bnbt_tag" + CAtomInt( ++ucTag ).toString( );

				strTag = CFG_GetString( strName, string( ) );
			}

			strName = "bnbt_tag" + CAtomInt( ucTagPrevious ).toString( );
			CFG_Delete( strName );
			CFG_Close( CFG_FILE );
			m_vecTags.clear( );
			m_vecTagsMouse.clear( );
			initTags( );

			return JS_ReturnToPage( pRequest, pResponse, TAGS_HTML );
		}

		if( pRequest->mapParams["submit_tag_form_button"] == STR_SUBMIT )
		{
			const unsigned char cucTagCount( (unsigned char)atoi( pRequest->mapParams["tag_count"].c_str( ) ) );

			unsigned char ucTagCountLimit = cucTagCount;

			string strName = string( );
			string strImageOn = string( );
			string strImageOff = string( );
	
			if( ucTagCountLimit == 0 )
				ucTagCountLimit = 1;

			for( unsigned char ucCount = 1; ucCount <= ucTagCountLimit; ucCount++ )
			{
				strName = pRequest->mapParams["bnbt_tag" + CAtomInt( ucCount ).toString( ) + "_name"];
				strImageOn = pRequest->mapParams["bnbt_tag" + CAtomInt( ucCount ).toString( ) + "_imageon"];
				strImageOff = pRequest->mapParams["bnbt_tag" + CAtomInt( ucCount ).toString( ) + "_imageoff"];

				if( strName.empty( ) )
					strName = "Tag" + CAtomInt( ucCount ).toString( );

				if( !strImageOn.empty( ) )
					strImageOn = "|" + strImageOn;

				if( !strImageOff.empty( ) )
					strImageOff = "|" + strImageOff;

				CFG_SetString( "bnbt_tag" + CAtomInt( ucCount ).toString( ), strName + strImageOn + strImageOff );
			}

			strName = pRequest->mapParams["bnbt_tag_name"];

			if( !strName.empty( ) )
			{
				strImageOn = pRequest->mapParams["bnbt_tag_imageon"];
				strImageOff = pRequest->mapParams["bnbt_tag_imageoff"];

				if( !strImageOn.empty( ) )
					strImageOn = "|" + strImageOn;

				if( !strImageOff.empty( ) )
					strImageOff = "|" + strImageOff;

				CFG_SetString( "bnbt_tag" + CAtomInt( cucTagCount + 1 ).toString( ), strName + strImageOn + strImageOff );
			}

			const string cstrMouseovers( UTIL_ToLower( pRequest->mapParams["bnbt_use_mouseovers"] ) );

			if( cstrMouseovers == "on" )
			{
				CFG_SetInt( "bnbt_use_mouseovers", 1);
				m_bUseMouseovers = true;
			}
			else
			{
				CFG_SetInt( "bnbt_use_mouseovers", 0);
				m_bUseMouseovers = false;
			}

			CFG_Close( CFG_FILE );
			m_vecTags.clear( );
			m_vecTagsMouse.clear( );
			initTags( );

			return JS_ReturnToPage( pRequest, pResponse, TAGS_HTML );
		}

		// display
		HTML_Common_Begin( pRequest, pResponse, gmapLANG_CFG["admin_page"], string( CSS_ADMIN ), string( ), NOT_INDEX, CODE_200 );

		// javascript
		pResponse->strContent += "<script type=\"text/javascript\">\n";
		pResponse->strContent += "<!--\n";
		pResponse->strContent += "function delete_tag_confirm( TAG )\n";
		pResponse->strContent += "{\n";
		pResponse->strContent += "var name=confirm(\"" + gmapLANG_CFG["delete"] + " bnbt_tag\" + TAG )\n";
		pResponse->strContent += "if (name==true)\n";
		pResponse->strContent += "{\n";
		pResponse->strContent += "window.location=\"" + RESPONSE_STR_TAGS_HTML + "?ap_deltag=\" + TAG\n";
		pResponse->strContent += "}\n";
		pResponse->strContent += "}\n\n";
		pResponse->strContent += "function admin_swap_tag( TAG ) {\n";
		pResponse->strContent += "window.location=\"" + RESPONSE_STR_TAGS_HTML + "?ap_swaptag=\" + TAG\n";
		pResponse->strContent += "}\n\n";
		pResponse->strContent += "//-->\n";
		pResponse->strContent += "</script>\n\n";

		// table
		pResponse->strContent += "<div class=\"admin_table\">\n";
		pResponse->strContent += "<form method=\"get\" name=\"admintags\" onSubmit=\"return confirm(\'" + gmapLANG_CFG["Submit"] + "\')\" action=\"" + RESPONSE_STR_TAGS_HTML + "\">\n";
		pResponse->strContent += "<table summary=\"xadmintags\">\n";
		pResponse->strContent += "<tr>\n<th colspan=6>Admin Tags</th>";
		pResponse->strContent += "<th colspan=2><label for=\"id_mouseover\">Enable Mouseover</label><input name=\"bnbt_use_mouseovers\" id=\"id_mouseover\" type=checkbox";
		
		if( CFG_GetInt( "bnbt_use_mouseovers", 0 ) == 0 ? false : true )
			pResponse->strContent += " checked";

		pResponse->strContent += "></th>\n</tr>\n";

		pResponse->strContent += "<tr>\n";
		pResponse->strContent += "<th>Key</th>\n";
		pResponse->strContent += "<th>Tag</th>\n";
		pResponse->strContent += "<th>Image On</th>\n";
		pResponse->strContent += "<th>Image On URL</th>\n";
		pResponse->strContent += "<th>Image Off</th>\n";
		pResponse->strContent += "<th>Image Off URL</th>\n";
		pResponse->strContent += "<th colspan=2>Action</th>\n";
		pResponse->strContent += "</tr>\n";

		unsigned char ucTag = 1;
		unsigned char ucTagPrevious = 0;

		string strTagPrevious = string( );

		string strName = "bnbt_tag" + CAtomInt( ucTag ).toString( );
		string strTag = CFG_GetString( strName, string( ) );

		string :: size_type iSplit = 0;
		string :: size_type iSplitMouse = 0;

		while( !strTag.empty( ) )
		{
			iSplit = strTag.find( "|" );
			iSplitMouse = strTag.find( "|", iSplit + 1 );

			ucTagPrevious = ucTag;
			strTagPrevious = CAtomInt( ucTagPrevious ).toString( );

			strName = "bnbt_tag" + CAtomInt( ++ucTag ).toString( );

			if( ucTag % 2 )
				pResponse->strContent += "<tr class=\"even\">\n";
			else
				pResponse->strContent += "<tr class=\"odd\">\n";

			if( iSplit == string :: npos ) 
			{
				if( iSplitMouse == string :: npos )
				{
					pResponse->strContent += "<td>bnbt_tag" + strTagPrevious + "</td>\n";
					pResponse->strContent += "<td><span style=\"display:none\"><label for=\"id_bnbt_tag" + strTagPrevious + "_name\">" + UTIL_RemoveHTML( strTag ) + " Name</label></span><input name=\"bnbt_tag" + strTagPrevious + "_name\" id=\"id_bnbt_tag" + strTagPrevious + "_name\" type=text size=25 value=\"" + strTag + "\"></td>\n";
					pResponse->strContent += "<td></td>\n";
					pResponse->strContent += "<td><span style=\"display:none\"><label for=\"id_bnbt_tag" + strTagPrevious + "_imageon\">" + UTIL_RemoveHTML( strTag ) + " Image On URL</label></span><input name=\"bnbt_tag" + strTagPrevious + "_imageon\" id=\"id_bnbt_tag" + strTagPrevious + "_imageon\" type=text size=40 value=\"\"></td>\n";
					pResponse->strContent += "<td></td>\n";
					pResponse->strContent += "<td><span style=\"display:none\"><label for=\"id_bnbt_tag" + strTagPrevious + "_imageoff\">" + UTIL_RemoveHTML( strTag ) + " Image Off URL</label></span><input name=\"bnbt_tag" + strTagPrevious + "_imageoff\" id=\"id_bnbt_tag" + strTagPrevious + "_imageoff\" type=text size=40 value=\"\"></td>\n";

					pResponse->strContent += "<td>";

					if( ucTag > 2 )
					{
						pResponse->strContent += Button_JS_Link( "up_bnbt_tag" + strTagPrevious, "Move Up", "admin_swap_tag(" + CAtomInt( ucTag - 2 ).toString( ) + ")" );
						pResponse->strContent += "<br>";
					}

					if( !CFG_GetString( "bnbt_tag" + CAtomInt( ucTag ).toString( ), string( ) ).empty( ) )
						pResponse->strContent += Button_JS_Link( "down_bnbt_tag" + strTagPrevious, "Move Down", "admin_swap_tag(" + strTagPrevious + ")" );

					pResponse->strContent += "</td>\n";

					pResponse->strContent += "<td>";
					pResponse->strContent += Button_JS_Link( "bnbt_tag" + strTagPrevious, gmapLANG_CFG["delete"], "delete_tag_confirm(" + strTagPrevious +  ")" );
					pResponse->strContent += "</td>\n";
				}
				else
				{
					pResponse->strContent += "<td>bnbt_tag" + strTagPrevious + "</td>\n";
					pResponse->strContent += "<td><span style=\"display:none\"><label for=\"id_bnbt_tag" + strTagPrevious + "_name\">" + UTIL_RemoveHTML( strTag.substr( 0, iSplitMouse ) ) + " Name</label></span><input name=\"bnbt_tag" + strTagPrevious + "_name\" id=\"id_bnbt_tag" + strTagPrevious + "_name\" type=text size=25 value=\"" + UTIL_RemoveHTML( strTag.substr( 0, iSplitMouse ) ) + "\"></td>\n";
					pResponse->strContent += "<td></td>\n";
					pResponse->strContent += "<td><span style=\"display:none\"><label for=\"id_bnbt_tag" + strTagPrevious + "_imageon\">" + UTIL_RemoveHTML( strTag.substr( 0, iSplitMouse ) ) + " Image On URL</label></span><input name=\"bnbt_tag" + strTagPrevious + "_imageon\" id=\"id_bnbt_tag" + strTagPrevious + "_imageon\" type=text size=40 value=\"\"></td>\n";
					pResponse->strContent += "<td><img title=\"" + strTag.substr( iSplitMouse + 1 ) + "\" name=\"name_bnbt_tag" + strTagPrevious + "_imageoff\" alt=\"[" + strTag.substr( iSplitMouse + 1 ) + "]\" src=\"" + strTag.substr( iSplitMouse + 1 ) + "\"></td>\n";

					pResponse->strContent += "<td>";

					if( ucTag > 2 )
					{
						pResponse->strContent += Button_JS_Link( "up_bnbt_tag" + strTagPrevious, "Move Up", "admin_swap_tag(" + CAtomInt( ucTag - 2 ).toString( ) + ")" );
						pResponse->strContent += "<br>";
					}

					if( !CFG_GetString( "bnbt_tag" + CAtomInt( ucTag ).toString( ), string( ) ).empty( ) )
						pResponse->strContent += Button_JS_Link( "down_bnbt_tag" + strTagPrevious, "Move Down", "admin_swap_tag(" + strTagPrevious + ")" );

					pResponse->strContent += "</td>\n";

					pResponse->strContent += "<td>";
					pResponse->strContent += Button_JS_Link( "bnbt_tag" + strTagPrevious, gmapLANG_CFG["delete"], "delete_tag_confirm(" + strTagPrevious +  ")" );
					pResponse->strContent += "</td>\n";
				}
			}
			else
			{
				if( iSplitMouse == string :: npos )
				{
					pResponse->strContent += "<td>bnbt_tag" + strTagPrevious + "</td>\n";
					pResponse->strContent += "<td><span style=\"display:none\"><label for=\"id_bnbt_tag" + strTagPrevious + "_name\">" + UTIL_RemoveHTML( strTag.substr( 0, iSplit ) ) + " Name</label></span><input name=\"bnbt_tag" + strTagPrevious + "_name\" id=\"id_bnbt_tag" + strTagPrevious + "_name\" type=text size=25 value=\"" + UTIL_RemoveHTML( strTag.substr( 0, iSplit ) ) + "\"></td>\n";
					pResponse->strContent += "<td><img title=\"" + strTag.substr( iSplit + 1 ) + "\" name=\"name_bnbt_tag" + strTagPrevious + "_imageon\" alt=\"[" + strTag.substr( iSplit + 1 ) + "]\" src=\"" + strTag.substr( iSplit + 1 ) + "\"></td>\n";
					pResponse->strContent += "<td><span style=\"display:none\"><label for=\"id_bnbt_tag" + strTagPrevious + "_imageon\">" + UTIL_RemoveHTML( strTag.substr( 0, iSplit ) ) + " Image On URL</label></span><input name=\"bnbt_tag" + strTagPrevious + "_imageon\" id=\"id_bnbt_tag" + strTagPrevious + "_imageon\" type=text size=40 value=\"" + UTIL_RemoveHTML( strTag.substr( iSplit + 1 ) ) + "\"></td>\n";
					pResponse->strContent += "<td></td>\n";
					pResponse->strContent += "<td><span style=\"display:none\"><label for=\"id_bnbt_tag" + strTagPrevious + "_imageoff\">" + UTIL_RemoveHTML( strTag.substr( 0, iSplit ) ) + " Image Off URL</label></span><input name=\"bnbt_tag" + strTagPrevious + "_imageoff\" id=\"id_bnbt_tag" + strTagPrevious + "_imageoff\" type=text size=40 value=\"\"></td>\n";

					pResponse->strContent += "<td>";

					if( ucTag > 2 )
					{
						pResponse->strContent += Button_JS_Link( "up_bnbt_tag" + strTagPrevious, "Move Up", "admin_swap_tag(" + CAtomInt( ucTag - 2 ).toString( ) + ")" );
						pResponse->strContent += "<br>";
					}

					if( !CFG_GetString( "bnbt_tag" + CAtomInt( ucTag ).toString( ), string( ) ).empty( ) )
						pResponse->strContent += Button_JS_Link( "down_bnbt_tag" + strTagPrevious, "Move Down", "admin_swap_tag(" + strTagPrevious + ")" );

					pResponse->strContent += "</td>\n";

					pResponse->strContent += "<td>";
					pResponse->strContent += Button_JS_Link( "bnbt_tag" + strTagPrevious, gmapLANG_CFG["delete"], "delete_tag_confirm(" + strTagPrevious +  ")" );
					pResponse->strContent += "</td>\n";
				}
				else
				{
					pResponse->strContent += "<td>bnbt_tag" + strTagPrevious + "</td>\n";
					pResponse->strContent += "<td><span style=\"display:none\"><label for=\"id_bnbt_tag" + strTagPrevious + "_name\">" + UTIL_RemoveHTML( strTag.substr( 0, iSplit ) ) + " Name</label></span><input name=\"bnbt_tag" + strTagPrevious + "_name\" id=\"id_bnbt_tag" + strTagPrevious + "_name\" type=text size=25 value=\"" + UTIL_RemoveHTML( strTag.substr( 0, iSplit ) ) + "\"></td>\n";
					pResponse->strContent += "<td><img title=\"" + strTag.substr( iSplit + 1, iSplitMouse - iSplit - 1 ) + "\" name=\"name_bnbt_tag" + strTagPrevious + "_imageon\" alt=\"[" + strTag.substr( iSplit + 1, iSplitMouse - iSplit - 1 ) + "]\" src=\"" + strTag.substr( iSplit + 1, iSplitMouse - iSplit - 1 ) + "\"></td>\n";
					pResponse->strContent += "<td><span style=\"display:none\"><label for=\"id_bnbt_tag" + strTagPrevious + "_imageon\">" + UTIL_RemoveHTML( strTag.substr( 0, iSplit ) ) + " Image On URL</label></span><input name=\"bnbt_tag" + strTagPrevious + "_imageon\" id=\"id_bnbt_tag" + strTagPrevious + "_imageon\" type=text size=40 value=\"" + UTIL_RemoveHTML( strTag.substr( iSplit + 1, iSplitMouse - iSplit - 1 ) ) + "\"></td>\n";
					pResponse->strContent += "<td><img title=\"" + strTag.substr( iSplitMouse + 1, iSplitMouse + iSplit ) + "\" name=\"name_bnbt_tag" + strTagPrevious + "_imageoff\" alt=\"[" + strTag.substr( iSplitMouse + 1 ) + "]\" src=\"" + strTag.substr( iSplitMouse + 1 ) + "\"></td>\n";
					pResponse->strContent += "<td><span style=\"display:none\"><label for=\"id_bnbt_tag" + strTagPrevious + "_imageoff\">" + UTIL_RemoveHTML( strTag.substr( 0, iSplit ) ) + " Image Off URL</label></span><input name=\"bnbt_tag" + strTagPrevious + "_imageoff\" id=\"id_bnbt_tag" + strTagPrevious + "_imageoff\" type=text size=40 value=\"" + UTIL_RemoveHTML( strTag.substr( iSplitMouse + 1 ) ) + "\"></td>\n";

					pResponse->strContent += "<td>";

					if( ucTag > 2 )
					{
						pResponse->strContent += Button_JS_Link( "up_bnbt_tag" + strTagPrevious, "Move Up", "admin_swap_tag(" + CAtomInt( ucTag - 2 ).toString( ) + ")" );
						pResponse->strContent += "<br>";
					}

					if( !CFG_GetString( "bnbt_tag" + CAtomInt( ucTag ).toString( ), string( ) ).empty( ) )
						pResponse->strContent += Button_JS_Link( "down_bnbt_tag" + strTagPrevious, "Move Down", "admin_swap_tag(" + strTagPrevious + ")" );

					pResponse->strContent += "</td>\n";

					pResponse->strContent += "<td>";
					pResponse->strContent += Button_JS_Link( "bnbt_tag" + strTagPrevious, gmapLANG_CFG["delete"], "delete_tag_confirm(" + strTagPrevious + ")" );
					pResponse->strContent += "</td>\n";
				}
			}   

			pResponse->strContent += "</tr>\n";
			strTag = CFG_GetString( strName, string( ) );
		}

		if( ( ucTagPrevious ) % 2 )
			pResponse->strContent += "<tr class=\"even\">\n";
		else
			pResponse->strContent += "<tr class=\"odd\">\n";

		pResponse->strContent += "<td>*bnbt_tag" + CAtomInt( ucTag ).toString( ) + "</td>\n";
		pResponse->strContent += "<td><span style=\"display:none\"><label for=\"id_bnbt_tag_name\">New Name</label></span><input name=\"bnbt_tag_name\" id=\"id_bnbt_tag_name\" type=text size=25></td>\n";
		pResponse->strContent += "<td></td>\n";
		pResponse->strContent += "<td><span style=\"display:none\"><label for=\"bnbt_tag_imageon\">New Image On URL</label></span><input name=\"bnbt_tag_imageon\" id=\"bnbt_tag_imageon\" type=text size=40></td>\n";
		pResponse->strContent += "<td></td>\n";
		pResponse->strContent += "<td><span style=\"display:none\"><label for=\"id_bnbt_tag_imageoff\">New Image Off URL</label></span><input name=\"bnbt_tag_imageoff\" id=\"id_bnbt_tag_imageoff\" type=text size=40></td>\n";

		pResponse->strContent += "<td class=\"number_red\">";
		pResponse->strContent += Button_Submit( "submit_tag_form", gmapLANG_CFG["Submit"] );
		pResponse->strContent += "</td>\n";

		pResponse->strContent += "<td class=\"number_red\">";
		pResponse->strContent += Button_Reset( "reset_tag_form", gmapLANG_CFG["reset"] );
		pResponse->strContent += "</td>\n";

		pResponse->strContent += "</tr>\n";

		pResponse->strContent += "</table>\n";
		pResponse->strContent += "<p><input name=\"tag_count\" type=hidden value=\"" + strTagPrevious + "\"></p>\n";
		pResponse->strContent += "</form>\n";
		pResponse->strContent += "</div>\n\n";

		// Output common HTML ending
		HTML_Common_End( pRequest, pResponse, btv, NOT_INDEX, string( CSS_ADMIN ) );
	}
	else
	{
		// not authorized
		HTML_Common_Begin(  pRequest, pResponse, gmapLANG_CFG["admin_page"], string( CSS_ADMIN ), string( ), NOT_INDEX, CODE_401 );

		// Output common HTML ending
		HTML_Common_End( pRequest, pResponse, btv, NOT_INDEX, string( CSS_ADMIN ) );
	}
}

void CTracker :: serverResponseXTorrent( struct request_t *pRequest, struct response_t *pResponse )
{
	// Set the start time
	const struct bnbttv btv( UTIL_CurrentTime( ) );

	// Verify that the IP is permitted to access the tracker
	if( m_ucIPBanMode != 0 )
		if( IsIPBanned( pRequest, pResponse, btv, gmapLANG_CFG["admin_page"], string( CSS_ADMIN ), NOT_INDEX ) )
			return;

	if( pRequest->user.ucAccess & ACCESS_ADMIN )
	{
		//
		// Swap announce address
		//

		const unsigned char cucSwapAnnounce( (unsigned char)atoi( pRequest->mapParams["ap_swapannounce"].c_str( ) ) );

		if( cucSwapAnnounce != 0 )
		{
			unsigned char ucAnnounceList = 1;
			string strKeyNext = string( );
			string strAnnounceNext = string( );

			string strKey = "xbnbt_announce_list" + CAtomInt( ucAnnounceList ).toString( );
			string strAnnounceUrl = CFG_GetString( strKey, string( ) );

			while( !strAnnounceUrl.empty( ) )
			{
				if( ucAnnounceList == cucSwapAnnounce )
				{
					strKeyNext = "xbnbt_announce_list" + CAtomInt( ucAnnounceList + 1 ).toString( );
					strAnnounceNext = CFG_GetString( strKeyNext, string( ) );

					CFG_SetString( strKey, strAnnounceNext );
					CFG_SetString( strKeyNext, strAnnounceUrl );

					ucAnnounceList++;
				}
				else
				{
					strKey = "xbnbt_announce_list" + CAtomInt( ucAnnounceList ).toString( );

					CFG_SetString( strKey, strAnnounceUrl );
				}

				strKey = "xbnbt_announce_list" + CAtomInt( ++ucAnnounceList ).toString( );

				strAnnounceUrl = CFG_GetString( strKey, string( ) );
			}

			CFG_Close( CFG_FILE );

			return JS_ReturnToPage( pRequest, pResponse, XTORRENT_HTML );
		}

		//
		// Delete announce address
		//

		const unsigned char cucDelAnnounce( (unsigned char)atoi( pRequest->mapParams["ap_delannounce"].c_str( ) ) );

		if( cucDelAnnounce != 0 )
		{
			unsigned char ucAnnounce = 1;
			unsigned char ucAnnouncePrevious = 0;

			string strName = "xbnbt_announce_list" + CAtomInt( ucAnnounce ).toString( );
			string strAnnounce = CFG_GetString( strName, string( ) );

			while( !strAnnounce.empty( ) )
			{
				if( ucAnnounce == cucDelAnnounce )
					CFG_Delete( strName );

				if( ucAnnounce > cucDelAnnounce )
				{
					strName = "xbnbt_announce_list" + CAtomInt( ucAnnouncePrevious ).toString( );
					CFG_SetString( strName, strAnnounce );
				}

				ucAnnouncePrevious = ucAnnounce;

				strName = "xbnbt_announce_list" + CAtomInt( ++ucAnnounce ).toString( );

				strAnnounce = CFG_GetString( strName, string( ) );
			}

			strName = "xbnbt_announce_list" + CAtomInt( ucAnnouncePrevious ).toString( );
			CFG_Delete( strName );
			CFG_Close( CFG_FILE );

			return JS_ReturnToPage( pRequest, pResponse, XTORRENT_HTML );
		}

		//
		// Submit the announce form
		//

		if( pRequest->mapParams["submit_announce_form_button"] == STR_SUBMIT )
		{
			const unsigned char cucAnnounceCount( (unsigned char)atoi( pRequest->mapParams["announce_count"].c_str( ) ) );

			unsigned char ucAnnounceCountLimit = cucAnnounceCount;

			string strAnnounceURL = pRequest->mapParams["bnbt_force_announce_url"];
	
			CFG_SetString( "bnbt_force_announce_url", strAnnounceURL );

			m_strForceAnnounceURL = strAnnounceURL;
			
			if( ucAnnounceCountLimit == 0 )
				ucAnnounceCountLimit = 1;

			for( unsigned char ucCount = 1; ucCount <= ucAnnounceCountLimit; ucCount++ )
			{
				strAnnounceURL = pRequest->mapParams["xbnbt_announce_list" + CAtomInt( ucCount ).toString( ) + "_url"];

				CFG_SetString( "xbnbt_announce_list" + CAtomInt( ucCount ).toString( ), strAnnounceURL );
			}

			strAnnounceURL = pRequest->mapParams["xbnbt_announce_list_url"];

			if( !strAnnounceURL.empty( ) )
				CFG_SetString( "xbnbt_announce_list" + CAtomInt( cucAnnounceCount + 1 ).toString( ), strAnnounceURL );

			const string cstrForceAnnounceDL( pRequest->mapParams["bnbt_force_announce_on_download"] );

			if( cstrForceAnnounceDL == "on" || cstrForceAnnounceDL == "On" )
			{
				CFG_SetInt( "bnbt_force_announce_on_download", 1);
				m_bForceAnnounceOnDL = true;
			}
			else
			{
				CFG_SetInt( "bnbt_force_announce_on_download", 0);
				m_bForceAnnounceOnDL = false;
			}

			const string cstrEnableAnnounceList( pRequest->mapParams["bnbt_announce_list_enable"] );

			if( cstrEnableAnnounceList == "on" || cstrEnableAnnounceList == "On" )
			{
				CFG_SetInt( "bnbt_announce_list_enable", 1);
				m_bEnableAnnounceList = true;
			}
			else
			{
				CFG_SetInt( "bnbt_announce_list_enable", 0);
				m_bEnableAnnounceList = false;
			}

			const string cstrForceAnnounceUL( pRequest->mapParams["bnbt_force_announce_on_upload"] );

			if( cstrForceAnnounceUL == "on" || cstrForceAnnounceUL == "On" )
			{
				CFG_SetInt( "bnbt_force_announce_on_upload", 1);
				m_bForceAnnounceOnUL = true;
			}
			else
			{
				CFG_SetInt( "bnbt_force_announce_on_upload", 0);
				m_bForceAnnounceOnUL = false;
			}

			CFG_Close( CFG_FILE );

			return JS_ReturnToPage( pRequest, pResponse, XTORRENT_HTML );
		}

		//
		// display the announce page
		//

		xtorrent.bForceAnnounceDL = CFG_GetInt( "bnbt_force_announce_on_download", 0 ) == 0 ? false : true;
		xtorrent.bForceAnnounceUL = CFG_GetInt( "bnbt_force_announce_on_upload", 0 ) == 0 ? false : true;
		xtorrent.bEnableAnnounceList = CFG_GetInt( "bnbt_announce_list_enable", 0 ) == 0 ? false : true;

		// Output common HTML head
		HTML_Common_Begin( pRequest, pResponse, gmapLANG_CFG["admin_page"], string( CSS_ADMIN ), string( ), NOT_INDEX, CODE_200 );

		// javascript
		pResponse->strContent += "<script type=\"text/javascript\">\n";
		pResponse->strContent += "<!--\n";
		pResponse->strContent += "function delete_announce_confirm( ANNOUNCE )\n";
		pResponse->strContent += "{\n";
		pResponse->strContent += "var name=confirm(\"" + gmapLANG_CFG["delete"] + " xbnbt_announce_list\" + ANNOUNCE )\n";
		pResponse->strContent += "if (name==true)\n";
		pResponse->strContent += "{\n";
		pResponse->strContent += "window.location=\"" + RESPONSE_STR_XTORRENT_HTML + "?ap_delannounce=\" + ANNOUNCE\n";
		pResponse->strContent += "}\n";
		pResponse->strContent += "}\n\n";
		pResponse->strContent += "function admin_swap_announce( ANNOUNCE ) {\n";
		pResponse->strContent += "window.location=\"" + RESPONSE_STR_XTORRENT_HTML + "?ap_swapannounce=\" + ANNOUNCE\n";
		pResponse->strContent += "}\n\n";
		pResponse->strContent += "//-->\n";
		pResponse->strContent += "</script>\n\n";

		// table
		pResponse->strContent += "<div class=\"admin_table\">\n";
		pResponse->strContent += "<form method=\"get\" name=\"adminxtorrent\" onSubmit=\"return confirm(\'" + gmapLANG_CFG["Submit"] + "\')\" action=\"" + RESPONSE_STR_XTORRENT_HTML + "\">\n";
		pResponse->strContent += "<table summary=\"adminxtorrent\">\n";
		pResponse->strContent += "<tr>\n";
		pResponse->strContent += "<th colspan=2>Admin XTorrent</th>";
		pResponse->strContent += "<th colspan=1><label for=\"id_bnbt_force_announce_on_download\">Force Announce DL</label><input name=\"bnbt_force_announce_on_download\" id=\"id_bnbt_force_announce_on_download\" type=checkbox";

		if( xtorrent.bForceAnnounceDL )
			pResponse->strContent += " checked";

		pResponse->strContent += "></th>\n";
		pResponse->strContent += "<th colspan=1><label for=\"id_bnbt_force_announce_on_upload\">Force Announce UL</label><input name=\"bnbt_force_announce_on_upload\" id=\"id_bnbt_force_announce_on_upload\" type=checkbox";

		if( xtorrent.bForceAnnounceUL )
			pResponse->strContent += " checked";

		pResponse->strContent += "></th>\n";
		pResponse->strContent += "</tr>\n";
		pResponse->strContent += "<tr>\n";
		pResponse->strContent += "<th>Key</th>\n";
		pResponse->strContent += "<th>Announce URL</th>\n";
		pResponse->strContent += "<th colspan=2>Action</th>\n";
		pResponse->strContent += "</tr>\n";

		unsigned char ucAnnounce = 1;
		unsigned char ucAnnouncePrevious = 0;
		string strAnnouncePrevious = string( );
		string strKey = "xbnbt_announce_list" + CAtomInt( ucAnnounce ).toString( );
		string strAnnounceUrl = CFG_GetString( strKey, string( ) );
		string strForceAnnounceUrl = CFG_GetString( "bnbt_force_announce_url", string( ) );

		pResponse->strContent += "<tr class=\"even\">\n";
		pResponse->strContent += "<td>bnbt_force_announce_url</td>\n";
		pResponse->strContent += "<td><span style=\"display:none\"><label for=\"id_bnbt_force_announce_url\">Force Announce URL</label></span><input name=\"bnbt_force_announce_url\" id=\"id_bnbt_force_announce_url\" type=text size=60 value=\"" + UTIL_RemoveHTML( strForceAnnounceUrl ) + "\"></td>\n";
		pResponse->strContent += "<td></td>\n";
		pResponse->strContent += "<td></td>\n";
		pResponse->strContent += "</tr>\n";

		if( ucAnnounce % 2 )
			pResponse->strContent += "<tr class=\"even\">\n";
		else
			pResponse->strContent += "<tr class=\"odd\">\n";

		pResponse->strContent += "<th colspan=2></th>";
		pResponse->strContent += "<th colspan=2><label for=\"id_bnbt_announce_list_enable\">Announce List</label><input name=\"bnbt_announce_list_enable\" id=\"id_bnbt_announce_list_enable\" type=checkbox";

		if( xtorrent.bEnableAnnounceList )
			pResponse->strContent += " checked";

		pResponse->strContent += "></th>\n";
		pResponse->strContent += "</tr>\n";

		while( !strAnnounceUrl.empty( ) )
		{
			ucAnnouncePrevious = ucAnnounce;
			strAnnouncePrevious = CAtomInt( ucAnnouncePrevious ).toString( );

			strKey = "xbnbt_announce_list" + CAtomInt( ++ucAnnounce ).toString( );

			if( ucAnnounce % 2 )
				pResponse->strContent += "<tr class=\"even\">\n";
			else
				pResponse->strContent += "<tr class=\"odd\">\n";
 
			pResponse->strContent += "<td>xbnbt_announce_list" + strAnnouncePrevious + "</td>\n";
			pResponse->strContent += "<td><span style=\"display:none\"><label for=\"id_xbnbt_announce_list" + strAnnouncePrevious + "_url\">Announce URL " + strAnnouncePrevious + "</label></span><input name=\"xbnbt_announce_list" + strAnnouncePrevious + "_url\" id=\"id_xbnbt_announce_list" + strAnnouncePrevious + "_url\" type=text size=60 value=\"" + UTIL_RemoveHTML( strAnnounceUrl ) + "\"></td>\n";
			pResponse->strContent += "<td>";

			if( ucAnnounce > 2 )
			{
				pResponse->strContent += Button_JS_Link( "up_xbnbt_announce_list" + strAnnouncePrevious, "Move Up", "admin_swap_announce(" + CAtomInt( ucAnnounce - 2 ).toString( ) + ")" );
				pResponse->strContent += "<br>";
			}

			if( !CFG_GetString( "xbnbt_announce_list" + CAtomInt( ucAnnounce ).toString( ), string( ) ).empty( ) )
				pResponse->strContent += Button_JS_Link( "down_xbnbt_announce_list" + strAnnouncePrevious, "Move Down", "admin_swap_announce(" + strAnnouncePrevious + ")" );

			pResponse->strContent += "</td>\n";
			pResponse->strContent += "<td>";
			pResponse->strContent += Button_JS_Link( "xbnbt_announce_list" + strAnnouncePrevious, gmapLANG_CFG["delete"], "delete_announce_confirm(" + strAnnouncePrevious +  ")" );
			pResponse->strContent += "</td>\n";
			pResponse->strContent += "</tr>\n";

			strAnnounceUrl = CFG_GetString( strKey, string( ) );
		}

		if( ( ucAnnouncePrevious ) % 2 )
			pResponse->strContent += "<tr class=\"even\">\n";
		else
			pResponse->strContent += "<tr class=\"odd\">\n";

		pResponse->strContent += "<td>*xbnbt_announce_list" + CAtomInt( ucAnnounce ).toString( ) + "</td>\n";
		pResponse->strContent += "<td><span style=\"display:none\"><label for=\"id_xbnbt_announce_list_url\">New Announce URL</label></span><input name=\"xbnbt_announce_list_url\" id=\"id_xbnbt_announce_list_url\" type=text size=60></td>\n";
		pResponse->strContent += "<td class=\"number_red\">";
		pResponse->strContent += Button_Submit( "submit_announce_form", gmapLANG_CFG["Submit"] );
		pResponse->strContent += "</td>\n";
		pResponse->strContent += "<td class=\"number_red\">";
		pResponse->strContent += Button_Reset( "reset_announce_form", gmapLANG_CFG["reset"] );
		pResponse->strContent += "</td>\n";
		pResponse->strContent += "</tr>\n";
		pResponse->strContent += "</table>\n";
		pResponse->strContent += "<p><input name=\"announce_count\" type=hidden value=\"" + strAnnouncePrevious + "\"></p>\n";
		pResponse->strContent += "</form>\n";
		pResponse->strContent += "</div>\n\n";

		// Output common HTML tail
		HTML_Common_End( pRequest, pResponse, btv, NOT_INDEX, string( CSS_ADMIN ) );
	}
	else
	{
		// not authorized

		// Output common HTML head
		HTML_Common_Begin(  pRequest, pResponse, gmapLANG_CFG["admin_page"], string( CSS_ADMIN ), string( ), NOT_INDEX, CODE_401 );

		// Output common HTML tail
		HTML_Common_End( pRequest, pResponse, btv, NOT_INDEX, string( CSS_ADMIN ) );
	}
}
